############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Tests many variations of dataset iteration callbacks with skipping.
'''

from accelerator import SkipDataset, SkipSlice


def synthesis(job, slices):
	# Chain of four, (2, 1, 3, 1) slices with data, only second one has column c.
	def mk_ds(name, previous, *data, **columns):
		dw = job.datasetwriter(name=name, columns=columns, previous=previous)
		w = dw.get_split_write_list()
		for v in data:
			w(v)
		return dw.finish()
	ds1 = mk_ds('1', None, (0, 0), (1, 10), a='int32', b='int32')
	ds2 = mk_ds('2', ds1, (2, 20, 200), a='int32', b='int32', c='int32')
	ds3 = mk_ds('3', ds2, (3, 30), (4, 40), (5, 50), a='int32', b='int32')
	ds4 = mk_ds('4', ds3, (6, 60), a='int32', b='int32')

	# Sanity check.
	assert list(ds4.iterate_chain(0, 'a')) == [0, 2, 3, 6]
	assert list(ds4.iterate_chain(0, 'b')) == [0, 20, 30, 60]
	assert list(ds4.iterate_chain(1, 'b')) == [10, 40]
	assert list(ds4.iterate_chain(None, 'b')) == [0, 10, 20, 30, 40, 50, 60]

	# Don't skip anything, just check pre_callback happens at the right times.
	# First with just ds.
	current = []
	current_expect = []
	expect = {
		'1': [],
		'2': [0, 1],
		'3': [2],
		'4': [3, 4, 5],
	}
	seen_ds = set()
	def chk_pre_ds(ds):
		assert ds.name not in seen_ds
		seen_ds.add(ds.name)
		current_expect.extend(expect[ds.name])
		assert current_expect == current
	for v in ds4.iterate_chain(None, 'a', pre_callback=chk_pre_ds):
		current.append(v)
	assert seen_ds == set('1234')

	# Then with ds and sliceno.
	current = []
	current_expect = []
	expect = {
		('1', 0): 0,
		('1', 1): 1,
		('2', 0): 2,
		('3', 0): 3,
		('3', 1): 4,
		('3', 2): 5,
		('4', 0): 6,
	}
	seen_ds = set()
	def chk_pre_ds_s(ds, sliceno):
		t = (ds.name, sliceno)
		assert t not in seen_ds
		seen_ds.add(t)
		assert current_expect == current
		if t in expect:
			current_expect.append(expect[t])
	for v in ds4.iterate_chain(None, 'a', pre_callback=chk_pre_ds_s):
		current.append(v)
	assert seen_ds == set.union(*(set((n, s) for s in range(slices)) for n in '1234'))

	# Same for post_callback
	# First with just ds.
	current = []
	current_expect = []
	expect = {
		'1': [0, 1],
		'2': [2],
		'3': [3, 4, 5],
		'4': [6],
	}
	seen_ds = set()
	def chk_post_ds(ds):
		assert ds.name not in seen_ds
		seen_ds.add(ds.name)
		current_expect.extend(expect[ds.name])
		assert current_expect == current
	for v in ds4.iterate_chain(None, 'a', post_callback=chk_post_ds):
		current.append(v)
	assert seen_ds == set('1234')

	# Then with ds and sliceno.
	current = []
	current_expect = []
	expect = {
		('1', 0): 0,
		('1', 1): 1,
		('2', 0): 2,
		('3', 0): 3,
		('3', 1): 4,
		('3', 2): 5,
		('4', 0): 6,
	}
	seen_ds = set()
	def chk_post_ds_s(ds, sliceno):
		t = (ds.name, sliceno)
		assert t not in seen_ds
		seen_ds.add(t)
		if t in expect:
			current_expect.append(expect[t])
			assert current_expect == current
	for v in ds4.iterate_chain(None, 'a', post_callback=chk_post_ds_s):
		current.append(v)
	assert seen_ds == set.union(*(set((n, s) for s in range(slices)) for n in '1234'))

	# And now both together. With sliceno on pre_callback, without on post_callback.
	current = []
	current_expect = []
	expect = {
		('1', 1): 0,
		('1', 2): 1,
		('2', 1): 2,
		('3', 1): 3,
		('3', 2): 4,
		('3', 3): 5,
		('4', 1): 6,
	}
	seen_pre = set()
	seen_post = set()
	def chk_both_pre(ds, sliceno):
		t = (ds.name, sliceno)
		assert t not in seen_pre
		seen_pre.add(t)
		if t in expect:
			current_expect.append(expect[t])
	def chk_both_post(ds):
		assert ds.name not in seen_post
		seen_post.add(ds.name)
		if slices == 3 and ds.name == '3':
			# ('3', 3) never happens, so fake it.
			current_expect.append(5)
		assert current_expect == current, '%s %r %r'%(ds,current_expect, current)
	for v in ds4.iterate_chain(None, 'a', pre_callback=chk_both_pre, post_callback=chk_both_post):
		current.append(v)
	assert seen_pre == set.union(*(set((n, s) for s in range(slices)) for n in '1234'))
	assert seen_post == set('1234')

	# They seem to work, let's test skipping.
	# Skip dataset 2. (SkipDataset works.)
	def skip_2(ds):
		if ds.name == '2':
			raise SkipDataset
	assert list(ds4.iterate_chain(0, 'a', pre_callback=skip_2)) == [0, 3, 6]

	# Neither ds4 (which we call) not ds1 (which we start on) have column c.
	# We end up keeping only ds2. (Skipping uniterable works.)
	def skip_without_c(ds):
		if 'c' not in ds.columns:
			raise SkipDataset
	assert list(ds4.iterate_chain(None, 'c', pre_callback=skip_without_c)) == [200]

	# Skip slice 0. (SkipSlice works.)
	def skip_slice_0(ds, sliceno):
		if sliceno == 0:
			raise SkipSlice
	assert list(ds4.iterate_chain(None, 'a', pre_callback=skip_slice_0)) == [1, 4, 5]

	# Skip slice 0 in ds1, skip all of ds3. (Combining SkipSlice and SkipDataset works.)
	def skip_1s0_and_2(ds, sliceno):
		if ds.name == '1' and sliceno == 0:
			raise SkipSlice
		if ds.name == '3':
			raise SkipDataset
	assert list(ds4.iterate_chain(None, 'b', pre_callback=skip_1s0_and_2)) == [10, 20, 60]

	# Skip ds3 after first slice. (SkipDataset works within dataset.)
	def skip_ds3_after_s0(ds, sliceno):
		if ds.name == '3' and sliceno == 1:
			raise SkipDataset
	assert list(ds4.iterate_chain(None, 'b', pre_callback=skip_ds3_after_s0)) == [0, 10, 20, 30, 60]

	# Stop after first slice of ds3.
	def stop_after_ds3_s0_pre(ds, sliceno):
		if ds.name == '3' and sliceno == 1:
			raise StopIteration
	assert list(ds4.iterate_chain(None, 'b', pre_callback=stop_after_ds3_s0_pre)) == [0, 10, 20, 30]

	# Same thing but by using post_callback.
	def stop_after_ds3_s0_post(ds, sliceno):
		if ds.name == '3':
			raise StopIteration
	assert list(ds4.iterate_chain(None, 'b', post_callback=stop_after_ds3_s0_post)) == [0, 10, 20, 30]

	# And without sliceno, so we get all of ds3.
	def stop_after_ds3_post(ds):
		if ds.name == '3':
			raise StopIteration
	assert list(ds4.iterate_chain(None, 'b', post_callback=stop_after_ds3_post)) == [0, 10, 20, 30, 40, 50]

	# And finally both callbacks on plain .iterate.
	def skip_s1(ds, sliceno):
		if sliceno == 1:
			raise SkipSlice
	current = []
	at_last_called = [0]
	def at_last(ds):
		assert ds == ds3
		assert current == [30, 50]
		at_last_called[0] += 1
	for v in ds3.iterate(None, 'b', pre_callback=skip_s1, post_callback=at_last):
		current.append(v)
	assert at_last_called == [1]
	assert current == [30, 50]



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that dataset_unroundrobin with trigger_column produces the
correct order and switches slice at the expected point.
'''

from accelerator import subjobs

def assert_slice(ds, sliceno, want):
	got = list(ds.iterate(sliceno))
	assert got == want, "slice %s in %s (from %s, trigger %s) gave\n\t%r,\nwanted\n\t%r" % (sliceno, ds, ds.job.params.datasets.source, ds.job.params.options.trigger_column, got, want,)

def synthesis(job, slices):
	unrr_values = [
		(0, 'a', 'A', 42),
		(0, 'a', 'A', 42),
		(0, 'a', 'A', 42),
		(1, 'a', 'A', 42),
		(1, 'b', 'A', 42),
		(1, 'b', 'B', 42),
		(1, 'b', 'C', 42),
		(2, 'b', 'D', 42),
		(2, 'c', 'E', 42),
	]

	# Use only slice 0 and 2, with 4 and 5 lines.
	dw = job.datasetwriter(name='0 and 2', columns=dict(a='int32', b='ascii', c='ascii', d='int32'))
	for sliceno in range(slices):
		dw.set_slice(sliceno)
		if sliceno == 0:
			dw.write_list(unrr_values[0])
			dw.write_list(unrr_values[2])
			dw.write_list(unrr_values[4])
			dw.write_list(unrr_values[6])
		if sliceno == 2:
			dw.write_list(unrr_values[1])
			dw.write_list(unrr_values[3])
			dw.write_list(unrr_values[5])
			dw.write_list(unrr_values[7])
			dw.write_list(unrr_values[8])
	src_ds = dw.finish()

	for col, split_point in (
		('a', 7,),
		('b', 4,),
		('c', 5,),
		('d', 9,),
	):
		ds_unrr = subjobs.build('dataset_unroundrobin', source=src_ds, trigger_column=col).dataset()
		assert_slice(ds_unrr, None, unrr_values)
		assert_slice(ds_unrr, 0, unrr_values[:split_point])
		assert_slice(ds_unrr, 2, unrr_values[split_point:])

	# Plain round robin.
	dw = job.datasetwriter(name='rr', columns=dict(a='int32', b='ascii', c='ascii', d='int32'))
	for sliceno in range(slices):
		dw.set_slice(sliceno)
		for v in unrr_values[sliceno::slices]:
			dw.write_list(v)
	src_ds = dw.finish()
	# The lines we expect in a slice
	def lines_for_slice(sliceno):
		if src_ds.lines[sliceno] == 0:
			return
		start_pos = sum(src_ds.lines[:sliceno])
		if start_pos == 0:
			trigger_v = object() # Not equal to anything else
		else:
			trigger_v = unrr_values[start_pos - 1][ix]
		cand_values = unrr_values[start_pos:]
		for v in cand_values[:src_ds.lines[sliceno]]:
			if v[ix] != trigger_v: # assumes values don't return to trigger_v
				yield v
		if v[ix] == trigger_v:
			# If we reached the end of "our" values without starting we will
			# not keep going until the next trigger change.
			return
		trigger_v = v[ix]
		for v in cand_values[src_ds.lines[sliceno]:]:
			if v[ix] != trigger_v:
				break
			yield v
	for ix, col in enumerate('abcd'):
		ds_unrr = subjobs.build('dataset_unroundrobin', source=src_ds, trigger_column=col).dataset()
		assert_slice(ds_unrr, None, unrr_values)
		for sliceno in range(slices):
			assert_slice(ds_unrr, sliceno, list(lines_for_slice(sliceno)))



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
from os import environ
from accelerator.build import JobList
from accelerator.job import Job
from accelerator.shell.parser import split_tildes, urd_call_w_tildes
from accelerator.error import UrdError
from accelerator.compat import url_quote


def main(argv, cfg):
	prog = argv.pop(0)
	user = environ.get('USER', 'NO-USER')
	if '--help' in argv or '-h' in argv:
		fh = sys.stdout if argv else sys.stderr
		print('usage: %s path [path [...]]' % (prog,), file=fh)
		print(file=fh)
		print('path is an optionally shortened path to an urd list.', file=fh)
		print('one element is a list name, two elements are list and timestamp.', file=fh)
		print('(and three elements are the whole thing, user/list/timestamp.)', file=fh)
		print(file=fh)
		print('use "path/since/ts" or just "path/" to list timestamps', file=fh)
		print('use "/" (or nothing) to list all lists', file=fh)
		print(file=fh)
		print('you can also use :urdlist:[entry] job specifiers. urdlist follows the same', file=fh)
		print('path rules as above, entry is an optional argument to joblist.get() printing', file=fh)
		print('just the resultant jobid.', file=fh)
		print(file=fh)
		print('in job specifiers you can use ~ to move to forwards (later, down) along a', file=fh)
		print('list of timestamps and ^ to move backwards (earlier, up).', file=fh)
		print(file=fh)
		print('examples:', file=fh)
		print('  "%s example" is "%s %s/example/latest"' % (prog, prog, user,), file=fh)
		print('  "%s :example:" is also "%s %s/example/latest"' % (prog, prog, user,), file=fh)
		print('  "%s example/2021-04-14" is "%s %s/example/2021-04-14"' % (prog, prog, user,), file=fh)
		print('  "%s :foo/bar/first:" is "%s foo/bar/first"' % (prog, prog,), file=fh)
		print('  "%s :bar/first~:" is the second timestamp in %s/bar' % (prog, user,), file=fh)
		print('  "%s example/" is "%s %s/example/since/0"' % (prog, prog, user,), file=fh)
		return not argv
	def resolve_path_part(path):
		if not path:
			return []
		if path == '/':
			return ['list']
		path = [url_quote(el) for el in path.split('/')]
		if path[-1] == '':
			path.pop()
			since = ['since', '0']
		elif len(path) > 2 and path[-2] == 'since':
			since = path[-2:]
			path = path[:-2]
		else:
			since = None
		if len(path) < 3 - bool(since):
			path.insert(0, user)
		if since:
			path.append(since[0])
			path.append(since[1] + '?captions')
		elif len(path) < 3:
			path.append('latest')
		return path
	def urd_get(path):
		if path.startswith(':'):
			a = path[1:].split(':', 1)
			if len(a) == 1:
				print('%r should have two or no :' % (path,), file=sys.stderr)
				return None, None
			path = a[0]
			try:
				entry = int(a[1], 10)
			except ValueError:
				entry = a[1] or None
			path, tildes = split_tildes(path)
		else:
			entry = tildes = None
		path = resolve_path_part(path)
		if len(path) != 3 and tildes:
			print("path %r isn't walkable (~^)" % ('/'.join(path),), file=sys.stderr)
			return None, None
		if len(path) != 3 and entry is not None:
			print("path %r doesn't take an entry (%r)" % ('/'.join(path), entry,), file=sys.stderr)
			return None, None
		try:
			res = urd_call_w_tildes(cfg, '/'.join(path), tildes)
		except UrdError as e:
			print(e, file=sys.stderr)
			res = None
		return res, entry
	for path in argv or ('/',):
		res, entry = urd_get(path)
		if not res:
			continue
		print(fmt(res, entry))

def fmt(res, entry):
	if not res:
		return ''
	def fmt_caption(path, caption):
		return template % (path, caption,) if caption else path
	if isinstance(res, list):
		if isinstance(res[0], list):
			tlen = max(len(ts) for ts, _ in res)
			template = '%%-%ds : %%s' % (tlen,)
			return '\n'.join(fmt_caption(*item) for item in res)
		else:
			return '\n'.join(res)
	joblist = JobList(Job(j, m) for m, j in res['joblist'])
	if entry:
		return joblist.get(entry, '')
	if res['deps']:
		deps = sorted(
			('%s/%s' % (k, v['timestamp'],), v['caption'],)
			for k, v in res['deps'].items()
		)
		if len(deps) > 1:
			plen = max(len(path) for path, _ in deps)
			template = '%%-%ds : %%s' % (plen,)
			deps = '\n           '.join(fmt_caption(*dep) for dep in deps)
		else:
			template = '%s : %s'
			deps = fmt_caption(*deps[0])
	else:
		deps = ''
	return "timestamp: %s\ncaption  : %s\ndeps     : %s\n%s" % (
		res['timestamp'],
		res['caption'],
		deps,
		joblist.pretty,
	)



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import division
from __future__ import absolute_import
from __future__ import print_function

from resource import getpagesize
from os import unlink
from os.path import exists
from mmap import mmap, PROT_READ
from shutil import copyfileobj
from struct import Struct
import itertools

from accelerator.compat import NoneType, unicode, imap, itervalues, PY2

from accelerator.extras import OptionEnum, DotDict
from accelerator.dsutil import typed_writer, typed_reader
from . import dataset_type

depend_extra = (dataset_type,)

description = r'''
Convert one or more columns in a dataset from bytes/ascii/unicode to any type.
Also rehashes if you type the hashlabel, or specify a new hashlabel.
You can also set hashlabel="" to not rehash (and get hashlabel=None).

Without filter_bad the method fails when a value fails to convert and
doesn't have a default. With filter_bad the value is filtered out
together with all other values on the same line.

With filter_bad, when rehashing or when typing a chain a new dataset is
produced. Any columns that do not have the same type over all the typed
datasets will be discarded in this case. You can set discard_untyped to
discard all untyped columns, or set it to False to get an error if any
columns were not preservable (except columns renamed over).
'''

TYPENAME = OptionEnum(dataset_type.convfuncs.keys())

options = {
	'column2type'               : {'COLNAME': TYPENAME},
	# Anything your system would accept in $TZ. Parses datetimes (but not
	# dates or times) as being in this timezone (you get UTC).
	# Doesn't work for %s (which is always in UTC).
	# No error checking can be done on this (tzset(3) can not return failure).
	# On most 32bit systems this will break dates outside about 1970 - 2037.
	# Setting this will mask most bad dates (mktime(3) "fixes" them).
	# Don't set this to 'UTC', leaving it as None is faster and safer.
	'timezone'                  : str,
	'hashlabel'                 : str, # leave as None to inherit hashlabel, set to '' to not have a hashlabel
	'defaults'                  : {}, # {'COLNAME': value}, unspecified -> method fails on unconvertible unless filter_bad
	'rename'                    : {}, # {'OLDNAME': 'NEWNAME'} doesn't shadow OLDNAME. (Other COLNAMEs use NEWNAME.) Use {'OLDNAME': None} to discard OLDNAME.
	'caption'                   : 'typed dataset',
	'discard_untyped'           : bool, # Make unconverted columns inaccessible ("new" dataset)
	'filter_bad'                : False, # Implies discard_untyped
	'numeric_comma'             : False, # floats as "3,14"
	'length'                    : -1, # Go back at most this many datasets. You almost always want -1 (which goes until previous.source)
	'as_chain'                  : False, # one dataset per slice if rehashing (avoids rewriting at the end)
	'compression'               : 6,     # gzip level
}

datasets = ('source', 'previous',)


byteslike_types = ('bytes', 'ascii', 'unicode',)

cstuff = dataset_type.init()

def prepare(job, slices):
	assert 1 <= options.compression <= 9
	if options.timezone:
		cstuff.backend.init(options.timezone.encode('utf-8'))
	else:
		cstuff.backend.init(cstuff.NULL)
	d = datasets.source
	chain = d.chain(stop_ds={datasets.previous: 'source'}, length=options.length)
	if len(chain) == 1:
		filename = d.filename
	else:
		filename = None
	lines = [sum(ds.lines[sliceno] for ds in chain) for sliceno in range(slices)]
	columns = {}
	column2type = dict(options.column2type)
	rev_rename = {}
	for k, v in options.rename.items():
		if k in d.columns and (v in column2type or options.discard_untyped is not True):
			if v in rev_rename:
				raise Exception('Both column %r and column %r rename to %r' % (rev_rename[v], k, v,))
			rev_rename[v] = k
	none_support = set()
	for colname, coltype in column2type.items():
		orig_colname = rev_rename.get(colname, colname)
		for ds in chain:
			if orig_colname not in ds.columns:
				raise Exception("Dataset %s doesn't have a column named %r (has %r)" % (ds, orig_colname, set(ds.columns),))
			if ds.columns[orig_colname].type not in byteslike_types:
				raise Exception("Dataset %s column %r is type %s, must be one of %r" % (ds, orig_colname, ds.columns[orig_colname].type, byteslike_types,))
		coltype = coltype.split(':', 1)[0]
		columns[colname] = dataset_type.typerename.get(coltype, coltype)
		if options.defaults.get(colname, False) is None:
			none_support.add(colname)
	if options.hashlabel is None:
		hashlabel_override = False
		hashlabel = options.rename.get(d.hashlabel, d.hashlabel)
		if hashlabel in columns:
			if rev_rename.get(hashlabel, hashlabel) != d.hashlabel:
				# hashlabel gets overwritten
				hashlabel = None
				hashlabel_override = True
		rehashing = (hashlabel in columns)
	else:
		hashlabel_override = True
		hashlabel = options.hashlabel or None
		rehashing = bool(hashlabel)
	if (options.filter_bad or rehashing or len(chain) > 1) and not options.discard_untyped:
		untyped_columns = set(d.columns)
		for ds in chain:
			untyped_columns &= set(ds.columns)
		orig_columns = set(rev_rename.get(n, n) for n in columns)
		untyped_columns -= set(columns) # anything renamed over is irrelevant
		untyped_columns -= orig_columns
		if options.discard_untyped is False:
			missing = set(d.columns) - untyped_columns - set(columns) - orig_columns
			if missing:
				raise Exception('discard_untyped is False, but not all columns in %s exist in the whole chain (missing %r)' % (d, missing,))
		unkeepable = set()
		for colname in sorted(untyped_columns):
			target_colname = options.rename.get(colname, colname)
			if target_colname is None or target_colname in columns:
				continue
			ts = set(ds.columns[colname].type for ds in chain)
			if len(ts) == 1:
				t = ts.pop()
				columns[target_colname] = t
				column2type[target_colname] = dataset_type.copy_types[t]
				if chain.none_support(colname):
					none_support.add(colname)
			else:
				# We could be a bit more generous and use bytes for other
				# workable combinations, or even unicode for {ascii, unicode}.
				unkeepable.add(colname)
		if options.discard_untyped is False:
			assert not unkeepable, 'The following columns have varying types: %r' % (unkeepable,)
	if options.filter_bad or rehashing or options.discard_untyped or len(chain) > 1:
		parent = None
	else:
		parent = datasets.source
	if hashlabel and hashlabel not in columns:
		if options.hashlabel:
			raise Exception("Can't rehash on discarded column %r." % (hashlabel,))
		hashlabel = None # it gets inherited from the parent if we're keeping it.
		hashlabel_override = False
	for colname in none_support:
		columns[colname] = (columns[colname], True)
	dws = []
	if rehashing:
		previous = datasets.previous
		if options.as_chain:
			# The last slice that actually has data in it becomes 'default'.
			# Or slice 0 if the whole source is empty (we must produce a ds).
			default_sliceno = 0
			for sliceno in range(slices - 1, -1, -1):
				if chain.lines(sliceno):
					default_sliceno = sliceno
					break
		else:
			default_sliceno = None
		for sliceno in range(slices):
			if sliceno == default_sliceno or chain.lines(sliceno):
				if sliceno == default_sliceno:
					name = 'default'
				else:
					name = str(sliceno)
				dw = job.datasetwriter(
					columns=columns,
					caption='%s (from slice %d)' % (options.caption, sliceno,),
					hashlabel=hashlabel,
					filename=filename,
					previous=previous,
					meta_only=True,
					name=name,
					for_single_slice=sliceno,
				)
				previous = dw
			else:
				dw = None
			dws.append(dw)
	if rehashing and options.as_chain:
		dw = None
	else:
		if parent:
			# Discard columns with rename set to None (unless already renamed over)
			discard = {colname: None for colname, t in options.rename.items() if t is None and colname not in columns and colname in parent.columns}
			if hashlabel in discard:
				hashlabel = None
				hashlabel_override = True
				assert options.hashlabel is None, "Can't hash on discarded column %r" % (options.hashlabel,)
			columns.update(discard)
		dw = job.datasetwriter(
			columns=columns,
			caption=options.caption,
			hashlabel=hashlabel,
			hashlabel_override=hashlabel_override,
			filename=filename,
			parent=parent,
			previous=datasets.previous,
			meta_only=True,
		)
	return dw, dws, lines, chain, column2type, sorted(columns), rev_rename


def map_init(vars, name, z='badmap_size'):
	if not vars.badmap_size:
		pagesize = getpagesize()
		line_count = vars.lines[vars.sliceno]
		vars.badmap_size = (line_count // 8 // pagesize + 1) * pagesize
		vars.slicemap_size = (line_count * 2 // pagesize + 1) * pagesize
	fh = open(name, 'w+b')
	vars.map_fhs.append(fh)
	fh.truncate(vars[z])
	return fh.fileno()


def analysis(sliceno, slices, prepare_res):
	if options.numeric_comma:
		try_locales = [
			'da_DK', 'nb_NO', 'nn_NO', 'sv_SE', 'fi_FI',
			'en_ZA', 'es_ES', 'es_MX', 'fr_FR', 'ru_RU',
			'de_DE', 'nl_NL', 'it_IT',
		]
		for localename in try_locales:
			localename = localename.encode('ascii')
			if not cstuff.backend.numeric_comma(localename):
				break
			if not cstuff.backend.numeric_comma(localename + b'.UTF-8'):
				break
		else:
			raise Exception("Failed to enable numeric_comma, please install at least one of the following locales: " + " ".join(try_locales))
	dw, dws, lines, chain, column2type, _, rev_rename = prepare_res
	if dws:
		dw = dws[sliceno]
		if not dw:
			dummy = [0] * slices
			return {}, dummy, dummy, {}, dummy
		rehashing = True
	else:
		rehashing = False
	vars = DotDict(
		sliceno=sliceno,
		slices=slices,
		known_line_count=0,
		badmap_size=0,
		badmap_fd=-1,
		slicemap_size=0,
		slicemap_fd=-1,
		map_fhs=[],
		res_bad_count={},
		res_default_count={},
		res_minmax={},
		first_lap=True,
		rehashing=rehashing,
		hash_lines=None,
		dw=dw,
		chain=chain,
		lines=lines,
		column2type=column2type,
		rev_rename=rev_rename,
	)
	if options.filter_bad:
		vars.badmap_fd = map_init(vars, 'badmap%d' % (sliceno,))
		bad_count, default_count, minmax = analysis_lap(vars)
		if sum(sum(c) for c in itervalues(bad_count)):
			vars.first_lap = False
			vars.res_bad_count = {}
			final_bad_count, default_count, minmax = analysis_lap(vars)
			final_bad_count = [max(c) for c in zip(*final_bad_count.values())]
		else:
			final_bad_count = [0] * slices
	else:
		bad_count, default_count, minmax = analysis_lap(vars)
		final_bad_count = [0] * slices
	for fh in vars.map_fhs:
		fh.close()
	if rehashing:
		unlink('slicemap%d' % (sliceno,))
	return bad_count, final_bad_count, default_count, minmax, vars.hash_lines


# In python3 indexing into bytes gives integers (b'a'[0] == 97),
# this gives the same behaviour on python2. (For use with mmap.)
class IntegerBytesWrapper(object):
	def __init__(self, inner):
		self.inner = inner
	def close(self):
		self.inner.close()
	def __getitem__(self, key):
		return ord(self.inner[key])
	def __setitem__(self, key, value):
		self.inner[key] = chr(value)

# But even in python3 we can only get int8 support for free,
# and slicemap needs int16.
class Int16BytesWrapper(object):
	_s = Struct('=H')
	def __init__(self, inner):
		self.inner = inner
	def close(self):
		self.inner.close()
	def __getitem__(self, key):
		return self._s.unpack_from(self.inner, key * 2)[0]
	def __setitem__(self, key, value):
		self._s.pack_into(self.inner, key * 2, value)
	def __iter__(self):
		if PY2:
			def it():
				for o in range(len(self.inner) // 2):
					yield self[o]
		else:
			def it():
				for v, in self._s.iter_unpack(self.inner):
					yield v
		return it()


def analysis_lap(vars):
	if vars.rehashing:
		if vars.first_lap:
			out_fn = 'hashtmp.%d' % (vars.sliceno,)
			colname = vars.rev_rename.get(vars.dw.hashlabel, vars.dw.hashlabel)
			coltype = vars.column2type[options.rename.get(colname, colname)]
			vars.rehashing = False
			real_coltype = one_column(vars, colname, coltype, [out_fn], True)
			vars.rehashing = True
			assert vars.res_bad_count[colname] == [0] # imlicitly has a default
			vars.slicemap_fd = map_init(vars, 'slicemap%d' % (vars.sliceno,), 'slicemap_size')
			slicemap = mmap(vars.slicemap_fd, vars.slicemap_size)
			slicemap = Int16BytesWrapper(slicemap)
			hash = typed_writer(real_coltype).hash
			slices = vars.slices
			vars.hash_lines = hash_lines = [0] * slices
			for ix, value in enumerate(typed_reader(real_coltype)(out_fn)):
				dest_slice = hash(value) % slices
				slicemap[ix] = dest_slice
				hash_lines[dest_slice] += 1
			unlink(out_fn)
	for colname, coltype in vars.column2type.items():
		if vars.rehashing:
			out_fns = [vars.dw.column_filename(colname, sliceno=s) for s in range(vars.slices)]
		else:
			out_fns = [vars.dw.column_filename(colname)]
		one_column(vars, vars.rev_rename.get(colname, colname), coltype, out_fns)
	return vars.res_bad_count, vars.res_default_count, vars.res_minmax


def one_column(vars, colname, coltype, out_fns, for_hasher=False):
	if for_hasher:
		record_bad = skip_bad = False
	elif vars.first_lap:
		record_bad = options.filter_bad
		skip_bad = False
	else:
		record_bad = 0
		skip_bad = options.filter_bad
	minmax_fn = 'minmax%d' % (vars.sliceno,)

	fmt = fmt_b = None
	is_null_converter = False
	if coltype in dataset_type.convfuncs:
		shorttype = coltype
		_, cfunc, pyfunc = dataset_type.convfuncs[coltype]
	elif coltype.startswith('null_'):
		shorttype = coltype
		pyfunc = False
		cfunc = True
		is_null_converter = True
	else:
		shorttype, fmt = coltype.split(':', 1)
		_, cfunc, pyfunc = dataset_type.convfuncs[shorttype + ':*']
	if cfunc:
		cfunc = shorttype.replace(':', '_')
	if pyfunc:
		tmp = pyfunc(coltype)
		if callable(tmp):
			pyfunc = tmp
			cfunc = None
		else:
			pyfunc = None
			cfunc, fmt, fmt_b = tmp
	if coltype == 'number':
		cfunc = 'number'
	elif coltype == 'number:int':
		coltype = 'number'
		cfunc = 'number'
		fmt = "int"
	assert cfunc or pyfunc, coltype + " didn't have cfunc or pyfunc"
	coltype = shorttype
	in_fns = []
	offsets = []
	max_counts = []
	for d in vars.chain:
		assert colname in d.columns, '%s not in %s' % (colname, d,)
		if not is_null_converter:
			assert d.columns[colname].type in byteslike_types, '%s has bad type in %s' % (colname, d,)
		in_fns.append(d.column_filename(colname, vars.sliceno))
		if d.columns[colname].offsets:
			offsets.append(d.columns[colname].offsets[vars.sliceno])
			max_counts.append(d.lines[vars.sliceno])
		else:
			offsets.append(0)
			max_counts.append(-1)
	if cfunc:
		default_value = options.defaults.get(colname, cstuff.NULL)
		if for_hasher and default_value is cstuff.NULL:
			if coltype.startswith('bits'):
				# No None-support.
				default_value = '0'
			else:
				default_value = None
		default_len = 0
		if default_value is None:
			default_value = cstuff.NULL
			default_value_is_None = True
		else:
			default_value_is_None = False
			if default_value != cstuff.NULL:
				if isinstance(default_value, unicode):
					default_value = default_value.encode("utf-8")
				default_len = len(default_value)
		c = getattr(cstuff.backend, 'convert_column_' + cfunc)
		if vars.rehashing:
			c_slices = vars.slices
		else:
			c_slices = 1
		bad_count = cstuff.mk_uint64(c_slices)
		default_count = cstuff.mk_uint64(c_slices)
		gzip_mode = "wb%d" % (options.compression,)
		safe_to_skip_write = vars.rehashing and not options.as_chain
		res = c(*cstuff.bytesargs(in_fns, len(in_fns), out_fns, gzip_mode, minmax_fn, default_value, default_len, default_value_is_None, fmt, fmt_b, record_bad, skip_bad, vars.badmap_fd, vars.badmap_size, c_slices, vars.slicemap_fd, vars.slicemap_size, bad_count, default_count, offsets, max_counts, safe_to_skip_write))
		assert not res, 'Failed to convert ' + colname
		vars.res_bad_count[colname] = list(bad_count)
		vars.res_default_count[colname] = sum(default_count)
		coltype = coltype.split(':', 1)[0]
		if is_null_converter:
			real_coltype = vars.chain[0].columns[colname].backing_type
			# Some lines may have been filtered out, so these minmax values
			# could be wrong. There's no easy/cheap way to fix that though,
			# and they will never be wrong in the bad direction.
			vars.res_minmax[colname] = [vars.chain.min(colname), vars.chain.max(colname)]
		else:
			real_coltype = dataset_type.typerename.get(coltype, coltype)
			if exists(minmax_fn):
				with typed_reader(real_coltype)(minmax_fn) as it:
					vars.res_minmax[colname] = list(it)
				unlink(minmax_fn)
	else:
		# python func
		if for_hasher:
			raise Exception("Can't hash on column of type %s." % (coltype,))
		nodefault = object()
		if colname in options.defaults:
			default_value = options.defaults[colname]
			if default_value is not None:
				if isinstance(default_value, unicode):
					default_value = default_value.encode('utf-8')
				default_value = pyfunc(default_value)
		else:
			default_value = nodefault
		if options.filter_bad:
			badmap = mmap(vars.badmap_fd, vars.badmap_size)
			if PY2:
				badmap = IntegerBytesWrapper(badmap)
		if vars.rehashing:
			slicemap = mmap(vars.slicemap_fd, vars.slicemap_size)
			slicemap = Int16BytesWrapper(slicemap)
			bad_count = [0] * vars.slices
		else:
			bad_count = [0]
			chosen_slice = 0
		default_count = 0
		dont_minmax_types = {'bytes', 'ascii', 'unicode', 'json', 'complex32', 'complex64'}
		real_coltype = dataset_type.typerename.get(coltype, coltype)
		do_minmax = real_coltype not in dont_minmax_types
		fhs = [typed_writer(real_coltype)(fn) for fn in out_fns]
		write = fhs[0].write
		col_min = col_max = None
		it = itertools.chain.from_iterable(d._column_iterator(vars.sliceno, colname, _type='bytes') for d in vars.chain)
		for ix, v in enumerate(it):
			if vars.rehashing:
				chosen_slice = slicemap[ix]
				write = fhs[chosen_slice].write
			if skip_bad:
				if badmap[ix // 8] & (1 << (ix % 8)):
					bad_count[chosen_slice] += 1
					continue
			try:
				v = pyfunc(v)
			except ValueError:
				if default_value is not nodefault:
					v = default_value
					default_count += 1
				elif record_bad:
					bad_count[chosen_slice] += 1
					bv = badmap[ix // 8]
					badmap[ix // 8] = bv | (1 << (ix % 8))
					continue
				else:
					raise Exception("Invalid value %r with no default in %s" % (v, colname,))
			if do_minmax and not isinstance(v, NoneType):
				if col_min is None:
					col_min = col_max = v
				if v < col_min: col_min = v
				if v > col_max: col_max = v
			write(v)
		for fh in fhs:
			fh.close()
		if vars.rehashing:
			slicemap.close()
		if options.filter_bad:
			badmap.close()
		vars.res_bad_count[colname] = bad_count
		vars.res_default_count[colname] = default_count
		vars.res_minmax[colname] = [col_min, col_max]
	return real_coltype

def synthesis(slices, analysis_res, prepare_res):
	dw, dws, lines, _, column2type, columns, rev_rename = prepare_res
	analysis_res = list(analysis_res)
	if options.filter_bad:
		bad_line_count_per_slice = [sum(data[1]) for data in analysis_res]
		lines = [num - b for num, b in zip(lines, bad_line_count_per_slice)]
		bad_line_count_total = sum(bad_line_count_per_slice)
		if bad_line_count_total:
			print('Slice   Bad line count')
			for sliceno, cnt in enumerate(bad_line_count_per_slice):
				print('%5d   %d' % (sliceno, cnt,))
			print('total   %d' % (bad_line_count_total,))
			print()
			print('Slice   Bad line number')
			reported_count = 0
			for sliceno, data in enumerate(analysis_res):
				if sum(data[1]) and reported_count < 32:
					with open('badmap%d' % (sliceno,), 'rb') as fh:
						badmap = mmap(fh.fileno(), 0, prot=PROT_READ)
						for ix, v in enumerate(imap(ord, badmap)):
							if v:
								for jx in range(8):
									if v & (1 << jx):
										print('%5d   %d' % (sliceno, ix * 8 + jx,))
										reported_count += 1
										if reported_count >= 32: break
								if reported_count >= 32: break
						badmap.close()
			if reported_count >= 32:
				print('...')
			print()
			print('Bad line count   Column')
			for colname in columns:
				cnt = sum(sum(data[0].get(colname, ())) for data in analysis_res)
				if cnt:
					print('%14d   %s' % (cnt, colname,))
			print()
		for sliceno in range(slices):
			unlink('badmap%d' % (sliceno,))
	if options.defaults and sum(sum(data[2].values()) for data in analysis_res):
		print('Defaulted values')
		for colname in sorted(options.defaults):
			defaulted = [data[2][colname] for data in analysis_res]
			if sum(defaulted):
				print('    %s:' % (colname,))
				print('        Slice   Defaulted line count')
				slicecnt = 0
				for sliceno, cnt in enumerate(defaulted):
					if cnt:
						print('        %5d   %d' % (sliceno, cnt,))
						slicecnt += 1
				if slicecnt > 1:
					print('        total   %d' % (sum(defaulted),))
	if dws: # rehashing
		if dw: # not as a chain
			final_bad_count = [data[1] for data in analysis_res]
			hash_lines = [data[4] for data in analysis_res]
			for colname in dw.columns:
				for sliceno in range(slices):
					out_fn = dw.column_filename(colname, sliceno=sliceno)
					with open(out_fn, 'wb') as out_fh:
						for s in range(slices):
							if hash_lines[s][sliceno] - final_bad_count[s][sliceno]:
								src_fn = dws[s].column_filename(colname, sliceno=sliceno)
								with open(src_fn, 'rb') as in_fh:
									copyfileobj(in_fh, out_fh)
			for sliced_dw in dws:
				if sliced_dw:
					sliced_dw.discard()
			for sliceno, counts in enumerate(zip(*[data[4] for data in analysis_res])):
				bad_counts = (data[1][sliceno] for data in analysis_res)
				dw.set_lines(sliceno, sum(counts) - sum(bad_counts))
			for sliceno, data in enumerate(analysis_res):
				dw.set_minmax(sliceno, data[3])
		else:
			for sliceno, data in enumerate(analysis_res):
				if dws[sliceno]:
					dws[sliceno].set_minmax(-1, data[3])
					for s, count in enumerate(data[4]):
						dws[sliceno].set_lines(s, count - data[1][s])
	else:
		for sliceno, count in enumerate(lines):
			dw.set_lines(sliceno, count)
		for sliceno, data in enumerate(analysis_res):
			dw.set_minmax(sliceno, data[3])
	used = {rev_rename.get(colname, colname) for colname in column2type}
	discarded = set(datasets.source.columns) - used
	if discarded:
		print('Discarded columns:')
		template = '    %%-%ds  %%s' % (max(len(colname) for colname in discarded),)
		for colname in discarded:
			print(template % (colname, datasets.source.columns[colname].type,))



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# In methods you only use the status function from this module.
# It works like this:
#
# with status("some text") as update:
#   long running code here
#
# then you can see how long your code has been doing that part in ^T.
# If you want to change the status text (without resetting the time,
# and possibly with other statuses nested below it) you can call the
# update function with the new text. If you don't need to update the
# text you can of course use just "with status(...):".
#
# Several built in functions will call this for you, notably dataset
# iterators and pickle load/save functions.

from __future__ import print_function
from __future__ import division

from contextlib import contextmanager
from functools import partial
from time import sleep
from traceback import print_exc
from threading import Lock
from weakref import WeakValueDictionary
import socket
import os

from accelerator.compat import str_types, iteritems, monotonic

from accelerator import g


status_tree = {}
status_all = WeakValueDictionary()
status_stacks_lock = Lock()


# all currently (or recently) running launch.py PIDs
class Children(set):
	def add(self, pid):
		with status_stacks_lock:
			set.add(self, pid)
	def remove(self, pid):
		with status_stacks_lock:
			d = status_all.pop(pid, None)
			if d and d.parent_pid in status_all:
				status_all[d.parent_pid].children.pop(pid, None)
			status_tree.pop(pid, None)
			set.remove(self, pid)
children = Children()


_cookie = 0

_local_status = []
_exc_status = (None, ())

@contextmanager
def status(msg):
	if g.running in ('server', 'build', 'shell',):
		yield lambda _: None
		return
	global _cookie
	_cookie += 1
	cookie = str(_cookie)
	t = str(monotonic())
	typ = 'push'
	# capture the PID here, because update might be called in a different process
	pid = os.getpid()
	update_local = _local_status.append
	def update(msg):
		assert msg and isinstance(msg, str_types) and '\0' not in msg
		update_local(msg)
		_send(typ, '\0'.join((msg, t, cookie)), pid=pid)
	update(msg)
	update_local = partial(_local_status.__setitem__, len(_local_status) - 1)
	typ = 'update'
	try:
		yield update
	except Exception as e:
		global _exc_status
		if _exc_status[0] != e:
			_exc_status = (e, list(_local_status))
		raise
	finally:
		_local_status.pop()
		_send('pop', cookie)

# Same interface as status, but without the actual reporting
@contextmanager
def dummy_status(msg):
	assert msg and isinstance(msg, str_types) and '\0' not in msg
	yield lambda _: None

def _start(msg, parent_pid, is_analysis=False):
	global _cookie
	if is_analysis:
		_cookie += 1
		analysis_cookie = str(_cookie)
	else:
		analysis_cookie = ''
	_send('start', '%d\0%s\0%s\0%f' % (parent_pid, analysis_cookie, msg, monotonic(),))
	def update(msg):
		_send('update', '%s\0\0%s' % (msg, analysis_cookie,))
	return update

def _end(pid=None):
	_send('end', '', pid=pid)

def _output(pid, msg):
	_send('output', '%f\0%s' % (monotonic(), msg,), pid=pid)

def _clear_output(pid):
	_send('output', '', pid=pid)


def status_stacks_export():
	res = []
	last = [None]
	current = None
	def fmt(tree, start_indent=0):
		for pid, d in sorted(iteritems(tree), key=lambda i: (i[1].stack or ((0,),))[0][0]):
			last[0] = d
			indent = start_indent
			for msg, t, _ in d.stack:
				res.append((pid, indent, msg, t))
				indent += 1
			fmt(d.children, indent)
			if d.output:
				res.append((pid, -1,) + d.output)
	try:
		with status_stacks_lock:
			fmt(status_tree)
		if last[0]:
			current = last[0].summary
			if len(last[0].stack) > 1 and not current[1].endswith('analysis'):
				msg, t, _ = last[0].stack[1]
				current = (current[0], '%s %s' % (current[1], msg,), t,)
	except Exception:
		print_exc()
		res.append((0, 0, 'ERROR', monotonic()))
	return res, current

def print_status_stacks(stacks=None):
	if stacks == None:
		stacks, _ = status_stacks_export()
	report_t = monotonic()
	for pid, indent, msg, t in stacks:
		if indent < 0:
			print("%6d TAIL OF OUTPUT: (%.1f seconds ago)" % (pid, report_t - t,))
			msgs = list(filter(None, msg.split('\n')))[-3:]
			for msg in msgs:
				print("       \x1b[32m" + msg + "\x1b[m")
		else:
			print("%6d STATUS: %s%s (%.1f seconds)" % (pid, "    " * indent, msg, report_t - t))


def _find(pid, cookie):
	stack = status_all[pid].stack
	# should normally be the last one, or at least close to it.
	for ix in range(len(stack) -1, -1, -1):
		if stack[ix][2] == cookie:
			return stack, ix
	return stack, None

def statmsg_sink(sock):
	from accelerator.extras import DotDict
	while True:
		data = None
		try:
			data = sock.recv(1500)
			typ, pid, msg = data.decode('utf-8').split('\0', 2)
			pid = int(pid)
			with status_stacks_lock:
				if typ == 'push':
					msg, t, cookie = msg.split('\0', 3)
					t = float(t)
					status_all[pid].stack.append((msg, t, cookie))
				elif typ == 'pop':
					stack, ix = _find(pid, msg)
					if ix == len(stack) - 1:
						stack.pop()
					else:
						print('POP OF WRONG STATUS: %d:%s (index %s of %d)' % (pid, msg, ix, len(stack)))
				elif typ == 'update':
					msg, _, cookie = msg.split('\0', 3)
					stack, ix = _find(pid, cookie)
					if ix is None:
						print('UPDATE TO UNKNOWN STATUS %d:%s: %s' % (pid, cookie, msg))
					else:
						stack[ix] = (msg, stack[ix][1], cookie)
				elif typ == 'output':
					if msg:
						t, msg = msg.split('\0', 1)
						t = float(t)
						status_all[pid].output = (msg, t,)
					else:
						status_all[pid].output = None
				elif typ == 'start':
					parent_pid, is_analysis, msg, t = msg.split('\0', 3)
					parent_pid = int(parent_pid)
					t = float(t)
					d = DotDict()
					d.parent_pid = parent_pid
					d.children   = {}
					d.stack      = [(msg, t, is_analysis or None)]
					d.summary    = (t, msg, t,)
					d.output     = None
					if parent_pid in status_all:
						if is_analysis:
							msg, parent_t, _ = status_all[parent_pid].stack[0]
							d.summary = (parent_t, msg + ' analysis', t,)
						status_all[parent_pid].children[pid] = d
					else:
						status_tree[pid] = d
					status_all[pid] = d
					del d
				elif typ == 'end':
					d = status_all.pop(pid, None)
					if d:
						if d.parent_pid in status_all:
							status_all[d.parent_pid].children.pop(pid, None)
						del d
					status_tree.pop(pid, None)
				else:
					print('UNKNOWN MESSAGE: %r' % (data,))
		except Exception:
			print('Failed to process %r:' % (data,))
			print_exc()


def statmsg_endwait(pid, timeout):
	"""Wait for pid to be removed from status_stacks (to send 'end')"""
	for _ in range(10):
		with status_stacks_lock:
			d = status_all.get(pid)
			if not d:
				return
		sleep(timeout / 10)


_send_sock = None

def _send(typ, message, pid=None):
	global _send_sock
	if not _send_sock:
		fd = int(os.getenv('BD_STATUS_FD'))
		_send_sock = socket.fromfd(fd, socket.AF_UNIX, socket.SOCK_DGRAM)
	header = ('%s\0%d\0' % (typ, pid or os.getpid(),)).encode('utf-8')
	message = message.encode('utf-8')
	if len(message) > 1450:
		message = message[:300] + b'\n....\n' + message[-1100:]
		# Make sure we don't have any partial characters.
		message = message.decode('utf-8', 'ignore').encode('utf-8')
	msg = header + message
	for ix in range(5):
		try:
			_send_sock.send(msg)
			return
		except socket.error as e:
			print('Failed to send statmsg (type %s, try %d): %s' % (typ, ix, e))
			sleep(0.1 + ix)



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

description = r'''
Verify that all output, including a final exception, is saved
'''

from accelerator import subjobs, JobError
from time import sleep

options = dict(
	inner=False,
)


def synthesis():
	lines = ['printing a bunch of lines, this is line %d.' % (n,) for n in range(150)]
	if options.inner:
		for s in lines:
			print(s)
		raise Exception('this is an exception, but nothing went wrong')
	else:
		try:
			subjobs.build('test_output_on_error', inner=True)
		except JobError as e:
			job = e.job
		else:
			raise Exception("test_output_on_error with inner=True didn't fail")
		# give the iowrapper some time to finish
		for attempt in range(25):
			got_lines = job.output().split('\n')
			if got_lines[:len(lines)] == lines:
				for line in got_lines:
					if line == 'Exception: this is an exception, but nothing went wrong':
						return
			# not yet, wait a little (total of 30s)
			if attempt > 1:
				print('Output from %s has not appeared yet, waiting more (%d).' % (job, attempt,))
			sleep(attempt / 10.0)
		raise Exception('Not all output from %s was saved in OUTPUT' % (job,))



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Verify that using dataset_type with a hashlabel gives the same result as
first typing and then rehashing for various hashlabel types, including
with renamed columns and bad lines.

Also verify that hashing discards untyped columns, and that parent
hashlabel is inherited or discarded as appropriate.
'''

from itertools import cycle
from datetime import date, time, datetime

from accelerator.compat import unicode
from accelerator import subjobs
from accelerator.extras import DotDict
from accelerator.dsutil import typed_writer

def synthesis(job, slices):
	# Test keeping untyped columns.
	dw = job.datasetwriter(name='a', columns={'a': 'unicode', 'b': ('bytes', True), 'c': ('ascii', True), 'd': ('number', True)})
	write = dw.get_split_write()
	write('A', None, None, None)
	write('a', b'b', 'c', 0)
	a = dw.finish()
	assert a.hashlabel == None
	typed_a = subjobs.build('dataset_type', options=dict(hashlabel='a', column2type={'a': 'ascii'}), datasets=dict(source=a)).dataset()
	assert typed_a.hashlabel == 'a'
	assert set(typed_a.iterate(None)) == {('A', None, None, None), ('a', b'b', 'c', 0)}, typed_a

	# Test hashing on a column not explicitly typed.
	dw = job.datasetwriter(name='b', columns={'a': 'unicode', 'b': 'ascii', 'c': 'bytes', 'd': 'unicode'}, previous=a)
	write = dw.get_split_write()
	write('A', 'B', b'C', '1')
	b = dw.finish()
	assert b.hashlabel == None
	typed_b = subjobs.build('dataset_type', options=dict(hashlabel='a', column2type={'b': 'ascii'}), datasets=dict(source=b)).dataset()
	assert typed_b.hashlabel == 'a'
	assert set(typed_b.iterate(None)) == {('a', 'b'), ('A', None), ('A', 'B')}, typed_b

	# Test renaming over the original hashlabel
	dw = job.datasetwriter(name='c', columns={'a': 'unicode', 'b': 'ascii', 'c': 'bytes', 'd': 'unicode'}, hashlabel='a')
	write = dw.get_split_write()
	write('\xe5', 'b', b'c', '0')
	c = dw.finish()
	assert c.hashlabel == 'a'
	typed_c = subjobs.build('dataset_type', options=dict(column2type={'a': 'ascii', 'd': 'number'}, rename={'c': 'a'}), datasets=dict(source=c)).dataset()
	assert typed_c.hashlabel == None
	assert list(typed_c.iterate(None)) == [('c', 'b', b'c', 0)], typed_c

	# Test using the original names but for different columns (keeping hashlabel under new name)
	dw = job.datasetwriter(name='d', columns={'a': 'unicode', 'b': 'ascii', 'c': 'bytes', 'd': 'unicode'}, hashlabel='a')
	write = dw.get_split_write()
	write('\xc5', 'B', B'C', '1')
	d = dw.finish()
	assert d.hashlabel == 'a'
	typed_d = subjobs.build('dataset_type', options=dict(column2type={'a': 'bytes', 'b': 'ascii', 'c': 'int32_10', 'd': 'bytes'}, rename={'b': 'a', 'c': 'b', 'd': 'c', 'a': 'd'}), datasets=dict(source=d)).dataset()
	assert typed_d.hashlabel == 'd'
	assert list(typed_d.iterate(None)) == [(b'B', 'C', 1, b'\xc3\x85')], typed_c

	# Test various types for hashing and discarding of bad lines.
	for hl in (None, 'a', 'b', 'c'):
		dw = job.datasetwriter(name='hashed on %s' % (hl,), columns={'a': 'unicode', 'b': 'unicode', 'c': 'unicode'}, hashlabel=hl)
		w = dw.get_split_write()
		for ix in range(1000):
			w(unicode(ix), '%d.%d' % (ix, ix % 5 == 0), ('{"a": %s}' if ix % 3 else '%d is bad') % (ix,))
		src_ds = dw.finish()
		assert src_ds.hashlabel == hl
		test(src_ds, dict(column2type={'a': 'int32_10', 'b': 'number:int'}, filter_bad=True), 800)
		test(src_ds, dict(column2type={'a': 'int64_10', 'b': 'number', 'c': 'json'}, filter_bad=True), 666)
		test(src_ds, dict(column2type={'a': 'floatint32ei', 'b': 'number:int', 'c': 'json'}, filter_bad=True), 533)
		test(src_ds, dict(column2type={'from_a': 'number', 'from_b': 'float64', 'from_c': 'ascii'}, rename=dict(a='from_a', b='from_b', c='from_c')), 1000)
		test(src_ds, dict(column2type={'c': 'bits32_16', 'a': 'float32', 'b': 'bytes'}, rename=dict(a='c', b='a', c='b')), 1000)

	# this doesn't test as many permutations, it's just to test more column types.
	dw = job.datasetwriter(name='more types')
	cols = {
		'floatbooli': cycle(['1.42 or so', '0 maybe', '1 (exactly)']),
		'datetime:%Y%m%d %H:%M': ['2019%02d%02d 17:%02d' % (t % 12 + 1, t % 28 + 1, t % 60) for t in range(1000)],
		'date:%Y%m%d': ['2019%02d%02d' % (t % 12 + 1, t % 28 + 1,) for t in range(1000)],
		'time:%H:%M': ['%02d:%02d' % (t // 60, t % 60) for t in range(1000)],
		'timei:%H:%M': ['%02d:%02d%c' % (t // 60, t % 60, chr(t % 26 + 65)) for t in range(1000)],
	}
	gens = []
	for coltype, gen in cols.items():
		dw.add(coltype.split(':')[0], 'ascii')
		gens.append(iter(gen))
	dw.add('half', 'bytes')
	gens.append(cycle([b'1', b'no']))
	w = dw.get_split_write()
	for _ in range(1000):
		w(*map(next, gens))
	src_ds = dw.finish()
	assert src_ds.hashlabel == None
	column2type = {t.split(':')[0]: t for t in cols}
	for hl in column2type:
		hashed = subjobs.build('dataset_type', options=dict(column2type=column2type, hashlabel=hl), datasets=dict(source=src_ds)).dataset()
		assert hashed.hashlabel == hl
		unhashed = subjobs.build('dataset_type', options=dict(column2type=column2type), datasets=dict(source=src_ds)).dataset()
		assert unhashed.hashlabel == None
		rehashed = subjobs.build('dataset_hashpart', options=dict(hashlabel=hl), datasets=dict(source=unhashed)).dataset()
		assert rehashed.hashlabel == hl
		assert hashed.lines == rehashed.lines
		assert sum(hashed.lines) == 1000
		assert set(hashed.columns.keys()) == set(unhashed.columns.keys()) == set(rehashed.columns.keys())
		# and again with a bad column
		column2type['half'] = 'float32'
		hashed = subjobs.build('dataset_type', options=dict(column2type=column2type, hashlabel=hl, filter_bad=True), datasets=dict(source=src_ds)).dataset()
		assert hashed.hashlabel == hl
		unhashed = subjobs.build('dataset_type', options=dict(column2type=column2type, filter_bad=True), datasets=dict(source=src_ds)).dataset()
		assert unhashed.hashlabel == None
		rehashed = subjobs.build('dataset_hashpart', options=dict(hashlabel=hl), datasets=dict(source=unhashed)).dataset()
		assert rehashed.hashlabel == hl
		del column2type['half']
		assert hashed.lines == rehashed.lines
		assert sum(hashed.lines) == 500
		assert set(hashed.columns.keys()) == set(unhashed.columns.keys()) == set(rehashed.columns.keys())

	# test rehashing on a column we don't type, over all types.
	dw = job.datasetwriter(name='rehash all types', columns={
		'2type'   : ('ascii', True),
		'ascii'   : ('ascii', True),
		'bits32'  : ('bits32', False),
		'bits64'  : ('bits64', False),
		'bool'    : ('bool', True),
		'bytes'   : ('bytes', True),
		'date'    : ('date', True),
		'datetime': ('datetime', True),
		'float32' : ('float32', True),
		'float64' : ('float64', True),
		'int32'   : ('int32', True),
		'int64'   : ('int64', True),
		'json'    : ('json', True),
		'number'  : ('number', True),
		'time'    : ('time', True),
		'unicode' : ('unicode', True),
	})
	write = dw.get_split_write()
	data = {
		'42': ('ascii string', 100, 1000,  True, b'bytes string', date(2019, 12, 11), datetime(2019, 12, 11, 20, 7, 21), 1.5, 0.00000001, 99, -11, {"a": "b"}, 1e100, time(20, 7, 21), 'unicode string'),
		None: (          None,   0,    0,  None,            None,               None,                              None, None,      None, None, None,     None,  None,            None,             None),
		'18': ('ASCII STRING', 111, 1111, False, b'BYTES STRING', date(1868,  1,  3), datetime(1868,  1,  3, 13, 14, 5), 2.5, -0.0000001, 67, -99, [42, ".."], 5e100, time(13, 14, 5), 'UNICODE STRING'),
	}
	write('42', *data['42'])
	write(None, *data[None])
	write('18', *data['18'])
	src_ds = dw.finish()
	data['None'] = data.pop(None)
	type2type = {
		'ascii'   : 'unicode:ascii',
		'bool'    : 'unicode:ascii',
		'date'    : 'unicode:ascii',
		'datetime': 'unicode:ascii',
		'time'    : 'unicode:ascii',
		'bits32'  : 'bits32_10',
		'bits64'  : 'bits64_10',
		'bytes'   : 'bytes',
		'float32' : 'float32',
		'float64' : 'float64',
		'int32'   : 'int32_10',
		'int64'   : 'int64_10',
		'number'  : 'number',
		'unicode' : 'unicode:ascii',
	}
	for hl, typeas in sorted(type2type.items()):
		ds = subjobs.build('dataset_type', column2type={'2type': typeas}, hashlabel=hl, source=src_ds).dataset()
		seen = set()
		hl_hash = typed_writer(hl).hash
		for sliceno in range(slices):
			for line in ds.iterate(sliceno, None):
				key = line[0] or None
				if isinstance(key, float):
					key = int(key)
				if isinstance(key, bytes):
					key = key.decode('ascii')
				else:
					key = unicode(key)
				assert data.get(key) == line[1:], "%s (hl %s) didn't have the right data for line %r" % (ds, hl, line[0],)
				hv = line[sorted(src_ds.columns).index(hl)]
				assert hl_hash(hv) % slices == sliceno, "%s (hl %s) didn't hash %r correctly" % (ds, hl, hv,)
				assert key not in seen, "%s (hl %s) repeated line %s" % (ds, hl, line[0],)
				seen.add(key)
		assert seen == {'42', 'None', '18'}, "%s didn't have all lines (%r)" % (ds, seen,)

def test(src_ds, opts, expect_lines):
	opts = DotDict(opts)
	def rename(colname):
		return opts.get('rename', {}).get(colname, colname)
	cols = set(opts.column2type)
	opts.discard_untyped = True
	msg = 'Testing with types %s' % (', '.join(v for k, v in sorted(opts.column2type.items())),)
	expect_hl = None
	if src_ds.hashlabel and opts.column2type.get(src_ds.hashlabel) == 'json':
		# json is not hashable, so we have to override the hashlabel to nothing in this case.
		opts.hashlabel = ''
		msg += ' (clearing hashlabel)'
	elif src_ds.hashlabel:
		expect_hl = rename(src_ds.hashlabel)
		if expect_hl in opts.column2type:
			msg += ' (hashed on %s)' % (opts.column2type[expect_hl],)
		else:
			expect_hl = None
			msg += ' (hashed on <untyped column>)'
	print(msg)
	just_typed = subjobs.build('dataset_type', options=opts, datasets=dict(source=src_ds)).dataset()
	assert just_typed.hashlabel == expect_hl, just_typed
	assert set(just_typed.columns) == cols, just_typed
	assert sum(just_typed.lines) == expect_lines, just_typed
	if rename(src_ds.hashlabel) not in opts.column2type or opts.get('hashlabel') == '':
		assert just_typed.hashlabel is None, just_typed
	else:
		assert just_typed.hashlabel == rename(src_ds.hashlabel), just_typed
	del opts.discard_untyped
	rev_rename = {v: k for k, v in opts.get('rename', {}).items()}
	discard = set(src_ds.columns) - set(rev_rename.get(n, n) for n in cols)
	if discard:
		d = opts.get('rename', {})
		d.update({k: None for k in discard})
		opts.rename = d
	for hashlabel in cols:
		if opts.column2type[hashlabel] == 'json':
			# not hashable
			continue
		opts['hashlabel'] = hashlabel
		print('%s rehashed on %s' % (msg, opts.column2type[hashlabel],))
		hashed_by_type = subjobs.build('dataset_type', options=opts, datasets=dict(source=src_ds)).dataset()
		assert hashed_by_type.hashlabel == hashlabel, hashed_by_type
		assert set(hashed_by_type.columns) == cols, hashed_by_type
		assert sum(hashed_by_type.lines) == expect_lines, hashed_by_type
		hashed_after = subjobs.build('dataset_hashpart', options=dict(hashlabel=hashlabel), datasets=dict(source=just_typed)).dataset()
		assert hashed_after.hashlabel == hashlabel, hashed_after
		assert set(hashed_after.columns) == cols, hashed_after
		assert sum(hashed_after.lines) == expect_lines, hashed_after
		if src_ds.hashlabel:
			# if src_ds has a hashlabel then just_typed will actually already be hashed, so hashed_after
			# will have been hashed twice and therefore have a different order than hashed_by_type.
			if rename(src_ds.hashlabel) == hashlabel:
				# These should be the same though.
				subjobs.build('test_compare_datasets', datasets=dict(a=hashed_by_type, b=just_typed))
			hashed_by_type = subjobs.build('dataset_sort', options=dict(sort_columns=rename('a')), datasets=dict(source=hashed_by_type))
			hashed_after = subjobs.build('dataset_sort', options=dict(sort_columns=rename('a')), datasets=dict(source=hashed_after))
		subjobs.build('test_compare_datasets', datasets=dict(a=hashed_by_type, b=hashed_after))



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Produce a summary that build_tests can put in the result directory.
'''

from accelerator.job import Job
from accelerator.build import fmttime

options = {'joblist': []}

def synthesis(job):
	joblist = [Job(j) for j in options.joblist]
	total = sum(j.params.exectime.total for j in joblist)
	with job.open('summary.html', 'w', encoding='utf-8') as fh:
		fh.write('<h2>Test built these jobs in ' + fmttime(total) + '</h2>\n')
		fh.write('<ol>\n')
		list_item = '<li><a href="/job/{job}" target="_blank">{job}</a> {job.method} {time}</li>\n'
		for j in joblist:
			fh.write(list_item.format(job=j, time=fmttime(j.params.exectime.total)))
		fh.write('</ol>\n')



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2019-2021 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# grep in a dataset(chain)

from __future__ import division, print_function

import sys
import re
from multiprocessing import Process, JoinableQueue
from itertools import chain, repeat
import errno
from os import write

from accelerator.compat import ArgumentParser
from accelerator.compat import unicode, izip, imap
from .parser import name2ds
from accelerator import g

def main(argv, cfg):
	usage = "%(prog)s [options] pattern ds [ds [...]] [column [column [...]]"
	parser = ArgumentParser(usage=usage, prog=argv.pop(0))
	parser.add_argument('-c', '--chain',        action='store_true', help="follow dataset chains", )
	parser.add_argument('-C', '--color',        action='store_true', help="color matched text", )
	parser.add_argument('-i', '--ignore-case',  action='store_true', help="case insensitive pattern", )
	parser.add_argument('-H', '--headers',      action='store_true', help="print column names before output (and on each change)", )
	parser.add_argument('-o', '--ordered',      action='store_true', help="output in order (one slice at a time)", )
	parser.add_argument('-g', '--grep',         action='append',     help="grep this column only, can be specified multiple times", metavar='COLUMN')
	parser.add_argument('-s', '--slice',        action='append',     help="grep this slice only, can be specified multiple times",  type=int)
	parser.add_argument('-t', '--separator', help="field separator (default tab)", default='\t')
	parser.add_argument('-D', '--show-dataset', action='store_true', help="show dataset on matching lines", )
	parser.add_argument('-S', '--show-sliceno', action='store_true', help="show sliceno on matching lines", )
	parser.add_argument('-L', '--show-lineno',  action='store_true', help="show lineno (per slice) on matching lines", )
	parser.add_argument('pattern')
	parser.add_argument('dataset', help='can be specified in the same ways as for "ax ds"')
	parser.add_argument('columns', nargs='*', default=[])
	args = parser.parse_intermixed_args(argv)

	pat_s = re.compile(args.pattern                , re.IGNORECASE if args.ignore_case else 0)
	pat_b = re.compile(args.pattern.encode('utf-8'), re.IGNORECASE if args.ignore_case else 0)
	datasets = [name2ds(cfg, args.dataset)]
	columns = []

	separator_s = args.separator
	separator_b = separator_s.encode('utf-8')

	for ds_or_col in args.columns:
		if columns:
			columns.append(ds_or_col)
		else:
			try:
				datasets.append(name2ds(cfg, ds_or_col))
			except Exception:
				columns.append(ds_or_col)

	if not datasets:
		parser.print_help(file=sys.stderr)
		return 1

	grep_columns = set(args.grep or ())
	if grep_columns == set(columns):
		grep_columns = None

	if args.slice:
		want_slices = []
		for s in args.slice:
			assert 0 <= s < g.slices, "Slice %d not available" % (s,)
			if s not in want_slices:
				want_slices.append(s)
	else:
		want_slices = list(range(g.slices))

	if args.chain:
		datasets = list(chain.from_iterable(ds.chain() for ds in datasets))

	if columns:
		bad = False
		for ds in datasets:
			missing = set(columns) - set(ds.columns)
			if missing:
				print('ERROR: %s does not have columns %r' % (ds, missing,), file=sys.stderr)
				bad = True
		if bad:
			return 1

	def grep(ds, sliceno):
		# Use bytes for everything if anything is bytes, str otherwise. (For speed.)
		if any(ds.columns[col].backing_type == 'bytes' for col in (grep_columns or columns or ds.columns)):
			def strbytes(v):
				return str(v).encode('utf-8', 'replace')
			def mk_iter(col):
				if ds.columns[col].backing_type in ('bytes', 'unicode', 'ascii',):
					it = ds._column_iterator(sliceno, col, _type='bytes')
					if ds.columns[col].none_support:
						it = (b'None' if v is None else v for v in it)
					return it
				else:
					return imap(strbytes, ds._column_iterator(sliceno, col))
			chk = pat_b.search
		else:
			def mk_iter(col):
				if ds.columns[col].backing_type in ('unicode', 'ascii',):
					it = ds._column_iterator(sliceno, col, _type='unicode')
					if not ds.columns[col].none_support:
						return it
				else:
					it = ds._column_iterator(sliceno, col)
				return imap(unicode, it)
			chk = pat_s.search
		def fmt(v):
			if not isinstance(v, (unicode, bytes)):
				v = str(v)
			if isinstance(v, unicode):
				v = v.encode('utf-8', 'replace')
			return v
		def color(item):
			pos = 0
			parts = []
			for m in pat_b.finditer(item):
				a, b = m.span()
				parts.extend((item[pos:a], b'\x1b[31m', item[a:b], b'\x1b[m'))
				pos = b
			parts.append(item[pos:])
			return b''.join(parts)
		prefix = []
		if args.show_dataset:
			prefix.append(ds.encode('utf-8'))
		if args.show_sliceno:
			prefix.append(str(sliceno).encode('utf-8'))
		prefix = tuple(prefix)
		def show(prefix, items):
			items = map(fmt, items)
			if args.color:
				items = map(color, items)
			# This will be atomic if the line is not too long
			# (at least up to PIPE_BUF bytes, should be at least 512).
			write(1, separator_b.join(prefix + tuple(items)) + b'\n')
		if grep_columns and grep_columns != set(columns or ds.columns):
			grep_iter = izip(*(mk_iter(col) for col in grep_columns))
			lines_iter = ds.iterate(sliceno, columns)
		else:
			grep_iter = repeat(None)
			lines_iter = izip(*(mk_iter(col) for col in (columns or sorted(ds.columns))))
		lines = izip(grep_iter, lines_iter)
		if args.show_lineno:
			for lineno, (grep_items, items) in enumerate(lines):
				if any(imap(chk, grep_items or items)):
					show(prefix + (str(lineno).encode('utf-8'),), items)
		else:
			for grep_items, items in lines:
				if any(imap(chk, grep_items or items)):
					show(prefix, items)

	def one_slice(sliceno, q, wait_for):
		try:
			if q:
				q.get()
			for ds in datasets:
				if ds in wait_for:
					q.task_done()
					q.get()
				grep(ds, sliceno)
		except KeyboardInterrupt:
			return
		except IOError as e:
			if e.errno == errno.EPIPE:
				return
			else:
				raise
		finally:
			# Make sure we are joinable
			try:
				q.task_done()
			except Exception:
				pass

	headers_prefix = []
	if args.show_dataset:
		headers_prefix.append('[DATASET]')
	if args.show_sliceno:
		headers_prefix.append('[SLICE]')
	if args.show_lineno:
		headers_prefix.append('[LINE]')

	headers = {}
	if args.headers:
		if columns:
			current_headers = columns
		else:
			current_headers = None
			for ds in datasets:
				candidate_headers = sorted(ds.columns)
				if candidate_headers != current_headers:
					headers[ds] = current_headers = candidate_headers
			current_headers = headers.pop(datasets[0])
		def show_headers(headers):
			print('\x1b[34m' + separator_s.join(headers_prefix + headers) + '\x1b[m')
		show_headers(current_headers)

	queues = []
	children = []
	if not args.ordered:
		q = None
		wait_for = set(headers)
		for sliceno in want_slices[1:]:
			if wait_for:
				q = JoinableQueue()
				q.put(None)
				queues.append(q)
			p = Process(
				target=one_slice,
				args=(sliceno, q, wait_for),
				name='slice-%d' % (sliceno,),
			)
			p.daemon = True
			p.start()
			children.append(p)
		want_slices = want_slices[:1]

	try:
		for ds in datasets:
			if ds in headers:
				for q in queues:
					q.join()
				show_headers(headers.pop(ds))
				for q in queues:
					q.put(None)
			for sliceno in want_slices:
				grep(ds, sliceno)
		for c in children:
			c.join()
	except KeyboardInterrupt:
		print()



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
# Modifications copyright (c) 2020 Anders Berkeman                         #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division

import os
import signal
import sys
from collections import defaultdict
from importlib import import_module
from traceback import format_tb, format_exception_only
from time import sleep
import json
import ctypes

from accelerator.job import CurrentJob, WORKDIRS
from accelerator.compat import pickle, iteritems, setproctitle, QueueEmpty
from accelerator.compat import getarglist, monotonic
from accelerator.extras import job_params, ResultIterMagic
from accelerator.build import JobError
from accelerator.lockfree_queue import LockFreeQueue
from accelerator import g
from accelerator import blob
from accelerator import statmsg
from accelerator import dataset
from accelerator import iowrapper


g_allesgut = False
_prof_fd = -1


g_always = {'running',}
assert set(n for n in dir(g) if not n.startswith("__")) == g_always, "Don't put anything in g.py"


# C libraries may not flush their stdio handles now that
# they are "files", so we do that after each stage.
clib = ctypes.cdll.LoadLibrary(None)
def c_fflush():
	clib.fflush(None)


def writeall(fd, data):
	while data:
		data = data[os.write(fd, data):]


def call_analysis(analysis_func, sliceno_, delayed_start, q, preserve_result, parent_pid, output_fds, **kw):
	try:
		q.make_writer()
		# tell iowrapper our PID, so our output goes to the right status stack.
		# (the pty is not quite a transparent transport ('\n' transforms into
		# '\r\n'), so we use a fairly human readable encoding.)
		writeall(output_fds[sliceno_], b'%16x' % (os.getpid(),))
		# use our iowrapper fd instead of stdout/stderr
		os.dup2(output_fds[sliceno_], 1)
		os.dup2(output_fds[sliceno_], 2)
		for fd in output_fds:
			os.close(fd)
		os.close(_prof_fd)
		slicename = 'analysis(%d)' % (sliceno_,)
		setproctitle(slicename)
		if delayed_start:
			os.close(delayed_start[1])
			update = statmsg._start('waiting for concurrency limit (%d)' % (sliceno_,), parent_pid, True)
			if os.read(delayed_start[0], 1) != b'a':
				raise Exception('bad delayed_start, giving up')
			update(slicename)
			os.close(delayed_start[0])
		else:
			statmsg._start(slicename, parent_pid, True)
		kw['sliceno'] = g.sliceno = sliceno_
		for dw in dataset._datasetwriters.values():
			if dw._for_single_slice is None:
				dw._set_slice(sliceno_)
		res = analysis_func(**kw)
		if preserve_result:
			# Remove defaultdicts until we find one with a picklable default_factory.
			# (This is what you end up doing manually anyway.)
			def picklable(v):
				try:
					pickle.dumps(v, pickle.HIGHEST_PROTOCOL)
					return True
				except Exception:
					return False
			def fixup(d):
				if isinstance(d, defaultdict) and not picklable(d.default_factory):
					if not d:
						return {}
					v = next(iteritems(d))
					if isinstance(v, defaultdict) and not picklable(v.default_factory):
						return {k: fixup(v) for k, v in iteritems(d)}
					else:
						return dict(d)
				else:
					return d
			def save(item, name):
				blob.save(fixup(item), name, sliceno=sliceno_, temp=True)
			if isinstance(res, tuple):
				if sliceno_ == 0:
					blob.save(len(res), "Analysis.tuple", temp=True)
				for ix, item in enumerate(res):
					save(item, "Analysis.%d." % (ix,))
			else:
				if sliceno_ == 0:
					blob.save(False, "Analysis.tuple", temp=True)
				save(res, "Analysis.")
		from accelerator.extras import saved_files
		dw_lens = {}
		dw_minmax = {}
		for name, dw in dataset._datasetwriters.items():
			if dw._for_single_slice in (None, sliceno_,):
				dw.close()
				dw_lens[name] = dw._lens
				dw_minmax[name] = dw._minmax
		c_fflush()
		q.put((sliceno_, monotonic(), saved_files, dw_lens, dw_minmax, None,))
		q.close()
	except:
		c_fflush()
		msg = fmt_tb(1)
		print(msg)
		q.put((sliceno_, monotonic(), {}, {}, {}, msg,))
		q.close()
		sleep(5) # give launcher time to report error (and kill us)
		exitfunction()

def fork_analysis(slices, concurrency, analysis_func, kw, preserve_result, output_fds, q):
	from multiprocessing import Process
	import gc
	children = []
	t = monotonic()
	pid = os.getpid()
	if hasattr(gc, 'freeze'):
		# See https://bugs.python.org/issue31558
		# (Though we keep the gc disabled by default.)
		gc.freeze()
	delayed_start = False
	delayed_start_todo = 0
	for i in range(slices):
		if i == concurrency:
			assert concurrency != 0
			# The rest will wait on this queue
			delayed_start = os.pipe()
			delayed_start_todo = slices - i
		p = Process(target=call_analysis, args=(analysis_func, i, delayed_start, q, preserve_result, pid, output_fds), kwargs=kw, name='analysis-%d' % (i,))
		p.start()
		children.append(p)
	for fd in output_fds:
		os.close(fd)
	if delayed_start:
		os.close(delayed_start[0])
	q.make_reader()
	per_slice = []
	temp_files = {}
	no_children_no_messages = False
	while len(per_slice) < slices:
		still_alive = []
		for p in children:
			if p.is_alive():
				still_alive.append(p)
			else:
				p.join()
				if p.exitcode:
					raise Exception("%s terminated with exitcode %d" % (p.name, p.exitcode,))
		children = still_alive
		# If a process dies badly we may never get a message here.
		# (iowrapper tries to tell us though.)
		# No need to handle that very quickly though, 10 seconds is fine.
		# (Typically this is caused by running out of memory.)
		try:
			msg = q.get(timeout=10)
			if not msg:
				# Notification from iowrapper, so we wake up (quickly) even if
				# the process died badly (e.g. from running out of memory).
				continue
			s_no, s_t, s_temp_files, s_dw_lens, s_dw_minmax, s_tb = msg
		except QueueEmpty:
			if not children:
				# No children left, so they must have all sent their messages.
				# Still, just to be sure there isn't a race, wait one iteration more.
				if no_children_no_messages:
					raise Exception("All analysis processes exited cleanly, but not all returned a result.")
				else:
					no_children_no_messages = True
			continue
		if s_tb:
			data = [{'analysis(%d)' % (s_no,): s_tb}, None]
			writeall(_prof_fd, json.dumps(data).encode('utf-8'))
			exitfunction()
		if delayed_start_todo:
			# Another analysis is allowed to run now
			os.write(delayed_start[1], b'a')
			delayed_start_todo -= 1
		per_slice.append((s_no, s_t))
		temp_files.update(s_temp_files)
		for name, lens in s_dw_lens.items():
			dataset._datasetwriters[name]._lens.update(lens)
		for name, minmax in s_dw_minmax.items():
			dataset._datasetwriters[name]._minmax.update(minmax)
	g.update_top_status("Waiting for all slices to finish cleanup")
	q.close()
	if delayed_start:
		os.close(delayed_start[1])
	for p in children:
		p.join()
	if preserve_result:
		res_seq = ResultIterMagic(slices, reuse_msg="analysis_res is an iterator, don't re-use it")
	else:
		res_seq = None
	return [v - t for k, v in sorted(per_slice)], temp_files, res_seq

def args_for(func):
	kw = {}
	for arg in getarglist(func):
		kw[arg] = getattr(g, arg)
	return kw


def fmt_tb(skip_level):
	msg = []
	e_type, e, tb = sys.exc_info()
	tb = format_tb(tb)[skip_level:]
	if isinstance(e, JobError):
		msg.append(e.format_msg())
		tb = tb[:-5] # the five innermost are in build.py and of no interest.
	msg.append("Traceback (most recent call last):\n")
	msg.extend(tb)
	from accelerator.statmsg import _exc_status
	if len(_exc_status[1]) > 1:
		msg.append("Status when the exception occurred:\n")
		for ix, txt in enumerate(_exc_status[1], 1):
			msg.append("  " * ix)
			msg.append(txt)
			msg.append("\n")
	msg.extend(format_exception_only(e_type, e))
	return ''.join(msg)


def execute_process(workdir, jobid, slices, concurrency, result_directory, common_directory, input_directory, index=None, workdirs=None, server_url=None, subjob_cookie=None, parent_pid=0):
	WORKDIRS.update(workdirs)

	g.job = jobid
	setproctitle('launch')
	path = os.path.join(workdir, jobid)
	try:
		os.chdir(path)
	except Exception:
		print("Cannot cd to workdir", path)
		exit(1)

	g.params = params = job_params()
	method_ref = import_module(params.package+'.a_'+params.method)
	g.sliceno = -1

	g.job = CurrentJob(jobid, params, result_directory, input_directory)
	g.slices = slices

	g.options          = params.options
	g.datasets         = params.datasets
	g.jobs             = params.jobs

	method_ref.options = params.options
	method_ref.datasets= params.datasets
	method_ref.jobs    = params.jobs

	g.server_url       = server_url
	g.running          = 'launch'
	statmsg._start('%s %s' % (jobid, params.method,), parent_pid)

	def dummy():
		pass

	prepare_func   = getattr(method_ref, 'prepare'  , dummy)
	analysis_func  = getattr(method_ref, 'analysis' , dummy)
	synthesis_func = getattr(method_ref, 'synthesis', dummy)

	synthesis_needs_analysis = 'analysis_res' in getarglist(synthesis_func)

	fd2pid, names, masters, slaves = iowrapper.setup(slices, prepare_func is not dummy, analysis_func is not dummy)
	def switch_output():
		fd = slaves.pop()
		os.dup2(fd, 1)
		os.dup2(fd, 2)
		os.close(fd)
	if analysis_func is dummy:
		q = None
	else:
		q = LockFreeQueue()
	iowrapper.run_reader(fd2pid, names, masters, slaves, q=q)
	for fd in masters:
		os.close(fd)

	# A chain must be finished from the back, so sort on that.
	sortnum_cache = {}
	def dw_sortnum(name):
		if name not in sortnum_cache:
			dw = dataset._datasetwriters.get(name)
			if not dw: # manually .finish()ed
				num = -1
			elif dw.previous and dw.previous.startswith(jobid + '/'):
				pname = dw.previous.split('/')[1]
				num = dw_sortnum(pname) + 1
			else:
				num = 0
			sortnum_cache[name] = num
		return sortnum_cache[name]

	prof = {}
	if prepare_func is dummy:
		prof['prepare'] = 0 # truthish!
	else:
		t = monotonic()
		switch_output()
		g.running = 'prepare'
		g.subjob_cookie = subjob_cookie
		setproctitle(g.running)
		with statmsg.status(g.running):
			g.prepare_res = method_ref.prepare(**args_for(method_ref.prepare))
			to_finish = [dw.name for dw in dataset._datasetwriters.values() if dw._started]
			if to_finish:
				with statmsg.status("Finishing datasets"):
					for name in sorted(to_finish, key=dw_sortnum):
						dataset._datasetwriters[name].finish()
		c_fflush()
		prof['prepare'] = monotonic() - t
	switch_output()
	setproctitle('launch')
	from accelerator.extras import saved_files
	if analysis_func is dummy:
		prof['per_slice'] = []
		prof['analysis'] = 0
	else:
		t = monotonic()
		g.running = 'analysis'
		g.subjob_cookie = None # subjobs are not allowed from analysis
		with statmsg.status('Waiting for all slices to finish analysis') as update:
			g.update_top_status = update
			prof['per_slice'], files, g.analysis_res = fork_analysis(slices, concurrency, analysis_func, args_for(analysis_func), synthesis_needs_analysis, slaves, q)
			del g.update_top_status
		prof['analysis'] = monotonic() - t
		saved_files.update(files)
	t = monotonic()
	g.running = 'synthesis'
	g.subjob_cookie = subjob_cookie
	setproctitle(g.running)
	with statmsg.status(g.running):
		synthesis_res = synthesis_func(**args_for(synthesis_func))
		if synthesis_res is not None:
			blob.save(synthesis_res, temp=False)
		if dataset._datasetwriters:
			with statmsg.status("Finishing datasets"):
				for name in sorted(dataset._datasetwriters, key=dw_sortnum):
					dataset._datasetwriters[name].finish()
	if dataset._datasets_written:
		with g.job.open('datasets.txt', 'w', encoding='utf-8') as fh:
			for name in dataset._datasets_written:
				fh.write(name)
				fh.write(u'\n')
	c_fflush()
	t = monotonic() - t
	prof['synthesis'] = t

	from accelerator.subjobs import _record
	return None, (prof, saved_files, _record)


def run(workdir, jobid, slices, concurrency, result_directory, common_directory, input_directory, index=None, workdirs=None, server_url=None, subjob_cookie=None, parent_pid=0, prof_fd=-1, debuggable=False):
	global g_allesgut, _prof_fd
	_prof_fd = prof_fd
	try:
		data = execute_process(workdir, jobid, slices, concurrency, result_directory, common_directory, input_directory, index=index, workdirs=workdirs, server_url=server_url, subjob_cookie=subjob_cookie, parent_pid=parent_pid)
		g_allesgut = True
	except Exception:
		msg = fmt_tb(2)
		print(msg)
		data = [{g.running: msg}, None]
		writeall(prof_fd, json.dumps(data).encode('utf-8'))
	finally:
		exitfunction()
	writeall(prof_fd, json.dumps(data).encode('utf-8'))


def exitfunction():
	if not g_allesgut:
		print('LAUNCH:  The deathening!')
		os.killpg(os.getpgid(0), signal.SIGKILL)



############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test that dataset_sort sorts stably.

Also tests DatasetWriter (again), and also DatasetWriter.finish()
'''

from accelerator.dataset import Dataset, DatasetWriter
from accelerator import subjobs

def prepare():
	dw = DatasetWriter()
	dw.add("str", "ascii")
	dw.add("num", "number")
	return dw

def analysis(sliceno, prepare_res):
	dw = prepare_res
	s = str(sliceno)
	# we expect to get a repeating str of range(slices) in "str" after sorting
	for ix in range(64):
		dw.write(s, ix)
	# less regular test case
	if sliceno == 0:
		dw.write("a", -1)
		dw.write("b", -1)
		dw.write("c", -2)
		dw.write("d", -1)
		dw.write("e", -1)
		dw.write("f", -1)
		dw.write("g", -2)
	if sliceno == 1:
		dw.write("h", -2)
		dw.write("i", -1)
		dw.write("j", -2)

def synthesis(params, prepare_res):
	dw = prepare_res
	source = dw.finish()
	jid = subjobs.build(
		"dataset_sort",
		options=dict(
			sort_columns="num",
			sort_across_slices=True,
		),
		datasets=dict(source=source),
	)
	ds = Dataset(jid)
	data = list(ds.iterate(None, "str"))
	good = list("cghjabdefi") + \
	       [str(sliceno) for sliceno in range(params.slices)] * 64
	assert data == good



############################################################################
#                                                                          #
# Copyright (c) 2017 eBay Inc.                                             #
# Modifications copyright (c) 2018-2021 Carl Drougge                       #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# a few things that differ between python2 and python3

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys
try:
	from setproctitle import setproctitle as _setproctitle, getproctitle
	# setproctitle init may be delayed until called (v1.2+) and should happen early
	getproctitle()
	del getproctitle
except ImportError:
	def _setproctitle(title): pass

if sys.version_info[0] == 2:
	PY2 = True
	PY3 = False
	import __builtin__ as builtins
	import cPickle as pickle
	FileNotFoundError = builtins.OSError
	from urllib import quote as url_quote, quote_plus, unquote_plus, urlencode
	from urllib2 import urlopen, Request, URLError, HTTPError
	from itertools import izip, izip_longest, imap, ifilter
	from Queue import Queue, Full as QueueFull, Empty as QueueEmpty
	try:
		from monotonic import monotonic
	except ImportError:
		from time import time as monotonic
	from types import NoneType
	str_types = (str, unicode,)
	int_types = (int, long,)
	num_types = (int, float, long,)
	unicode = builtins.unicode
	long = builtins.long
	def iterkeys(d):
		return d.iterkeys()
	def itervalues(d):
		return d.itervalues()
	def iteritems(d):
		return d.iteritems()
	from io import open
	def getarglist(func):
		from inspect import getargspec
		return getargspec(func).args
	def terminal_size():
		from termios import TIOCGWINSZ
		import struct
		from fcntl import ioctl
		from collections import namedtuple
		from os import environ
		def ifgood(name):
			try:
				v = int(environ[name])
				if v > 0:
					return v
			except (KeyError, ValueError):
				pass
		lines, columns = ifgood('LINES'), ifgood('COLUMNS')
		if not lines or not columns:
			try:
				fb_lines, fb_columns, _, _ = struct.unpack('HHHH', ioctl(0, TIOCGWINSZ, struct.pack('HHHH', 0, 0, 0, 0)))
			except Exception:
				fb_lines, fb_columns = 24, 80
		return namedtuple('terminal_size', 'columns lines')(columns or fb_columns, lines or fb_lines)
else:
	PY2 = False
	PY3 = True
	import builtins
	FileNotFoundError = builtins.FileNotFoundError
	import pickle
	from urllib.parse import quote as url_quote, quote_plus, unquote_plus, urlencode
	from urllib.request import urlopen, Request
	from urllib.error import URLError, HTTPError
	izip = zip
	from itertools import zip_longest as izip_longest
	imap = map
	ifilter = filter
	from queue import Queue, Full as QueueFull, Empty as QueueEmpty
	from time import monotonic
	NoneType = type(None)
	str_types = (str,)
	int_types = (int,)
	num_types = (int, float,)
	unicode = str
	open = builtins.open
	long = int
	def iterkeys(d):
		return iter(d.keys())
	def itervalues(d):
		return iter(d.values())
	def iteritems(d):
		return iter(d.items())
	def getarglist(func):
		from inspect import getfullargspec
		return getfullargspec(func).args
	from shutil import get_terminal_size as terminal_size

def first_value(d):
	return next(itervalues(d) if isinstance(d, dict) else iter(d))

def uni(s):
	if s is None:
		return None
	if isinstance(s, bytes):
		try:
			return s.decode('utf-8')
		except UnicodeDecodeError:
			return s.decode('iso-8859-1')
	return unicode(s)

# This is used in the method launcher to set different titles for each
# phase/slice. You can use it in the method to override that if you want.
def setproctitle(title):
	from accelerator import g
	if hasattr(g, 'params'):
		title = '%s %s (%s)' % (g.job, uni(title), g.params.method,)
	elif hasattr(g, 'job'):
		title = '%s %s' % (g.job, uni(title),)
	else:
		title = uni(title)
	if PY2:
		title = title.encode('utf-8')
	_setproctitle(title)

# allow_abbrev is 3.5+. it's not even available in the pypi backport of argparse.
# it also regrettably disables -abc for -a -b -c until 3.8.
from argparse import ArgumentParser as _ArgumentParser
if PY2:
	ArgumentParser = _ArgumentParser
else:
	class ArgumentParser(_ArgumentParser):
		def __init__(self, *a, **kw):
			return _ArgumentParser.__init__(self, *a, allow_abbrev=False, **kw)

# parse_intermixed_args is new in 3.7
if not hasattr(ArgumentParser, 'parse_intermixed_args'):
	ArgumentParser.parse_intermixed_args = ArgumentParser.parse_args



############################################################################
#                                                                          #
# Copyright (c) 2021 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

# This provides some of the functionality of multiprocessing.Queue but
# without using any locks, instead relying on writes up to PIPE_BUF being
# atomic. No threads are used.
#
# Only the creating process can .get(). .get() can take a timeout.
# Only child processes can .put(). .put() blocks until the pipe will take
# all the data.
# A dying writer will never corrupt or deadlock the queue, but can cause a
# memory leak in the reading process (until the queue object is destroyed).
# Large messages are probably quite inefficient.
#
# In short, you often can't use this. And when you do, you have to be careful.

from __future__ import print_function
from __future__ import division

import os
import sys
import select
import fcntl
import struct
import pickle
import errno
from accelerator.compat import QueueEmpty, monotonic


assert select.PIPE_BUF >= 512, "POSIX says PIPE_BUF is at least 512, you have %d" % (select.PIPE_BUF,)

PIPE_BUF = min(select.PIPE_BUF, 65536)
MAX_PART = PIPE_BUF - 6


def _nb(fd):
	fl = fcntl.fcntl(fd, fcntl.F_GETFL)
	fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)


class LockFreeQueue:
	def __init__(self):
		self.r, self.w = os.pipe()
		_nb(self.r)
		_nb(self.w)
		self._buf = b''
		self._pid = os.getpid()
		self._partial = {}

	def make_writer(self):
		os.close(self.r)
		self.r = -1

	def make_reader(self):
		os.close(self.w)
		self.w = -1

	def close(self):
		if self.r != -1:
			os.close(self.r)
			self.r = -1
		if self.w != -1:
			os.close(self.w)
			self.w = -1

	__del__ = close

	def get(self, block=True, timeout=0):
		assert os.getpid() == self._pid, "can only .get in creating process"
		assert self.w == -1, "call make_reader first"
		if timeout:
			deadline = monotonic() + timeout
		need_data = False
		eof = False
		while True:
			if not eof:
				try:
					data = os.read(self.r, PIPE_BUF)
					if not data:
						eof = True
					self._buf += data
					need_data = False
				except OSError:
					pass
			if len(self._buf) >= 6:
				z, pid = struct.unpack('<HI', self._buf[:6])
				assert pid, "all is lost"
				if len(self._buf) < 6 + z:
					need_data = True
				else:
					data = self._buf[6:6 + z]
					self._buf = self._buf[6 + z:]
					if pid not in self._partial:
						want_len = struct.unpack("<I", data[:4])[0]
						have = [data[4:]]
						have_len = len(have[0])
					else:
						want_len, have, have_len = self._partial[pid]
						have.append(data)
						have_len += len(data)
					if have_len == want_len:
						self._partial.pop(pid, None)
						data = b''.join(have)
						return pickle.loads(data)
					else:
						self._partial[pid] = (want_len, have, have_len)
			if len(self._buf) < 6 or need_data:
				if eof:
					raise QueueEmpty()
				if not block:
					if not select.select([self.r], [], [], 0)[0]:
						raise QueueEmpty()
				elif timeout:
					time_left = deadline - monotonic()
					if time_left <= 0:
						raise QueueEmpty()
					select.select([self.r], [], [], time_left)
				else:
					select.select([self.r], [], [])

	def put(self, msg):
		pid = os.getpid()
		assert pid != self._pid, "can only .put in other processes"
		assert self.r == -1, "call make_writer first"
		msg = pickle.dumps(msg, pickle.HIGHEST_PROTOCOL)
		msg = struct.pack('<I', len(msg)) + msg
		offset = 0
		while offset < len(msg):
			part = msg[offset:offset + MAX_PART]
			part = struct.pack('<HI', len(part), pid) + part
			offset += MAX_PART
			while True:
				select.select([], [self.w], [])
				try:
					wlen = os.write(self.w, part)
				except OSError as e:
					if e.errno == errno.EAGAIN:
						wlen = 0
					else:
						raise
				if wlen:
					if wlen != len(part):
						print("OS violates PIPE_BUF guarantees, all is lost.", file=sys.stderr)
						while True:
							try:
								# this should eventually make the other side read PID 0
								os.write(self.w, b'\0')
							except OSError:
								pass
					break

	def try_notify(self):
		pid = os.getpid()
		msg = struct.pack('<HII', 6, pid, 2) + b'N.' # pickled None
		try:
			os.write(self.w, msg)
			return True
		except OSError:
			return False



############################################################################
#                                                                          #
# Copyright (c) 2020 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test dataset_sort with trigger_column
'''

from operator import itemgetter

from accelerator import subjobs

def sort(src, ix, **kw):
	ds = subjobs.build('dataset_sort', source=src, sort_across_slices=True, **kw).dataset()
	want = sorted(src.iterate(None), key=itemgetter(ix))
	assert list(ds.iterate(None)) == want, '%s != sorted(%s)' % (ds, src,)
	return ds

def synthesis(job):
	dw = job.datasetwriter()
	dw.add('a', 'int32')
	dw.add('b', 'int32')
	dw.add('c', 'int32')
	w = dw.get_split_write()
	for ix in range(1, 11):
		w(1, 2, ix * 10 + 5)
		w(1, 2, ix * 10 + 3)
		w(1, 2, ix * 10 + 4)
		w(1, 3, ix * 1000 + 1)
		w(1, 3, ix * 1000 + 2)
		w(1, 3, ix * 1000 + 0)
	src = dw.finish()
	# Unchanging trigger
	a = sort(src, 0, sort_columns='a', trigger_column='a')
	assert set(a.lines) == {0, 60}, '%s %r' % (a, a.lines,)
	# Just two trigger values
	b = sort(src, 1, sort_columns='b', trigger_column='b')
	assert set(b.lines) == {0, 30}, '%s %r' % (b, b.lines,)
	# Trigger value changes every time - trigger_column should affect nothing
	c = sort(src, 2, sort_columns='c')
	ct = sort(src, 2, sort_columns='c', trigger_column='c')
	assert c.lines == ct.lines, '%s %r != %s %r' % (c, c.lines, ct, ct.lines,)
	# check that using second sort column as trigger works
	bc = sort(src, 2, sort_columns=['c', 'b'], trigger_column='b')
	assert set(bc.lines) == {0, 30}, '%s %r' % (bc, bc.lines,)



############################################################################
#                                                                          #
# Copyright (c) 2019-2021 Carl Drougge                                     #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import os
from select import select
from multiprocessing import Process
from subprocess import Popen
from time import sleep
import signal
from pty import openpty
import errno

from accelerator.workarounds import nonblocking
from accelerator.compat import setproctitle
from accelerator import statmsg


def main():
	# Run cat between ourselves and the real terminal, so we can set our
	# output (this pipe) non-blocking without messing up the real stdout.
	a, b = os.pipe()
	os.environ['BD_TERM_FD'] = str(b)
	Popen(['cat'], close_fds=True, stdin=a)
	# Use a pty for programs under us, so they stay line buffered.
	a, b = openpty()
	run_reader({}, None, [a], [b], 'main iowrapper.reader', None, True)
	os.dup2(b, 1)
	os.dup2(b, 2)
	os.close(a)
	os.close(b)


def setup(slices, include_prepare, include_analysis):
	os.mkdir('OUTPUT')
	names = []
	masters = []
	slaves = []
	def mk(name):
		a, b = openpty()
		masters.append(a)
		slaves.append(b)
		names.append(name)
	# order matters here, non-analysis outputs are pop()ed from the end.
	if include_analysis:
		for sliceno in range(slices):
			mk(str(sliceno))
	else:
		slices = 0 # for fd2pid later
	mk('synthesis')
	if include_prepare:
		mk('prepare')
	fd2pid = dict.fromkeys(masters[slices:], os.getpid())
	return fd2pid, names, masters, slaves


def run_reader(fd2pid, names, masters, slaves, process_name='iowrapper.reader', basedir='OUTPUT', is_main=False, q=None):
	syncpipe_r, syncpipe_w = os.pipe()
	args = (fd2pid, names, masters, slaves, process_name, basedir, is_main, syncpipe_r, syncpipe_w, q)
	p = Process(target=reader, args=args, name=process_name)
	p.start()
	if not is_main:
		os.close(int(os.environ['BD_TERM_FD']))
		del os.environ['BD_TERM_FD']
	os.close(syncpipe_w)
	# wait until reader is in a separate process group, so we don't kill it later
	os.read(syncpipe_r, 1)
	os.close(syncpipe_r)


MAX_OUTPUT = 640

def reader(fd2pid, names, masters, slaves, process_name, basedir, is_main, syncpipe_r, syncpipe_w, q):
	# don't get killed when we kill the job (will exit on EOF, so no output is lost)
	os.setpgrp()
	# we are safe now, the main process can continue
	os.close(syncpipe_w)
	os.close(syncpipe_r)
	signal.signal(signal.SIGTERM, signal.SIG_IGN)
	signal.signal(signal.SIGINT, signal.SIG_IGN)
	setproctitle(process_name)
	out_fd = int(os.environ['BD_TERM_FD'])
	for fd in slaves:
		os.close(fd)
	if q:
		q.make_writer()
	fd2fd = {}
	if not is_main:
		os.chdir(basedir)
		fd2name = dict(zip(masters, names))
		outputs = dict.fromkeys(masters, b'')
		if len(fd2pid) == 2:
			status_blacklist = set(fd2pid.values())
			assert len(status_blacklist) == 1, "fd2pid should only map to 1 value initially: %r" % (fd2pid,)
		else:
			status_blacklist = ()
			assert len(fd2pid) == 1, "fd2pid should have 1 or 2 elements initially"
	missed = [False]
	output_happened = False
	def try_print(data=b'\n\x1b[31m*** Some output not printed ***\x1b[m\n'):
		try:
			os.write(out_fd, data)
		except OSError:
			missed[0] = True
	# set output nonblocking, so we can't be blocked by terminal io.
	# errors generated here go to stderr, which is the real stderr
	# in the main iowrapper (so it can block) and goes to the main
	# iowrapper in the method iowrappers (so it can still block, but
	# is unlikely to do so for long).
	with nonblocking(out_fd):
		while masters:
			if missed[0]:
				# Some output failed to print last time around.
				# Wait up to one second for new data and then try
				# to write a message about that (before the new data).
				ready, _, _ = select(masters, [], [], 1.0)
				missed[0] = False
				try_print()
			else:
				ready, _, _ = select(masters, [], [])
			for fd in ready:
				try:
					data = os.read(fd, 65536)
				except OSError as e:
					# On Linux a pty will return
					# OSError: [Errno 5] Input/output error
					# instead of b'' for EOF. Don't know why.
					# Let's try to be a little restrictive in what we catch.
					if e.errno != errno.EIO:
						raise
					data = b''
				if data:
					if not is_main:
						if fd not in fd2pid:
							fd2pid[fd] = int(data[:16], 16)
							data = data[16:]
							if not data:
								continue
						if fd not in fd2fd:
							fd2fd[fd] = os.open(fd2name[fd], os.O_CREAT | os.O_WRONLY, 0o666)
						os.write(fd2fd[fd], data)
					try_print(data)
					output_happened = True
					if not is_main:
						outputs[fd] = (outputs[fd] + data[-MAX_OUTPUT:])[-MAX_OUTPUT:]
						statmsg._output(fd2pid[fd], outputs[fd].decode('utf-8', 'replace'))
				else:
					if q:
						# let fork_analysis know it's time to wake up
						# (in case the process died badly and didn't put anything in q)
						q.try_notify()
					if fd in fd2fd:
						os.close(fd2fd[fd])
						del fd2fd[fd]
					masters.remove(fd)
					os.close(fd)
					if not is_main:
						try:
							pid = fd2pid.pop(fd)
							if pid in status_blacklist:
								# don't do it for prepare as synthesis has the same PID.
								status_blacklist.remove(pid)
								# but clear the output if needed.
								if outputs[fd]:
									statmsg._clear_output(pid)
							else:
								statmsg._end(pid=pid)
						except Exception:
							# Failure can happen here if the method exits
							# before analysis (fd2pid not fully populated).
							pass
		if missed[0]:
			missed[0] = False
			try_print()
			if missed[0]:
				# Give it a little time, then give up.
				sleep(0.03)
				missed[0] = False
				try_print()
	if not output_happened and not is_main:
		os.chdir('..')
		os.rmdir(basedir)



# -*- coding: utf-8 -*-

############################################################################
#                                                                          #
# Copyright (c) 2019 Carl Drougge                                          #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License");          #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#  http://www.apache.org/licenses/LICENSE-2.0                              #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
#                                                                          #
############################################################################

from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

description = r'''
Test OptionEnum construction and enforcement.
'''

from accelerator.extras import OptionEnum
from accelerator.compat import PY2
from accelerator import subjobs
from accelerator import blob

abcd = OptionEnum("a b c d", none_ok=True)
efgh = OptionEnum(chr(n) for n in range(101, 105))
ijkl = OptionEnum(list("ijkl"), none_ok=True)
mnSTARop = OptionEnum("m n* o p")
qSTARrSTARst = OptionEnum("q* r* s t")
uni = OptionEnum("b l ä")

options = dict(
	abcd=abcd,
	efgh=efgh.g,
	ijkl=ijkl["k"],
	mnSTARop=mnSTARop.m,
	qSTARrSTARst=qSTARrSTARst["quu"],
	uni=uni.b,
	dict=dict(ijkl=ijkl),
	list=[mnSTARop.o],
	inner=False,
)

defaults = dict(options)
defaults["abcd"] = None
defaults["dict"] = {}

def check(**options):
	pass_options = dict(options)
	pass_options["inner"] = True
	want_res = dict(defaults)
	want_res.update(pass_options)
	jid = subjobs.build("test_optionenum", options=pass_options)
	res = blob.load(jobid=jid)
	assert res == want_res, "%r != %r from %r" % (res, want_res, options,)

def check_unbuildable(**options):
	try:
		check(**options)
	except Exception as e:
		if e.args[0].startswith("Submit failed"):
			return
		raise
	raise Exception("Building with options = %r should have failed but didn't" % (options,))

def synthesis():
	if options.inner:
		return dict(options)
	check(abcd="a", efgh="h")
	check(ijkl="j", efgh="e")
	if PY2:
		# Passing "ä" is fine on PY2 too, but we get UTF-8 byte strings back.
		check(uni=b"\xc3\xa4")
	else:
		check(uni="ä")
	check(ijkl=None, dict=dict(foo="j"))
	check(mnSTARop="p", qSTARrSTARst="qwe", ijkl="k")
	check(mnSTARop="nah", qSTARrSTARst="really good value\n")
	check(list=["noo"])
	check_unbuildable(abcd="A")
	check_unbuildable(efgh="a")
	check_unbuildable(qSTARrSTARst="sdf")
	check_unbuildable(mnSTARop=None)
	check_unbuildable(ijkl=None, dict=dict(foo="b"))
	check_unbuildable(list=["moo"])
